# -*- coding: utf-8 -*-
#   Copyright 2020 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.libs.math_constantmath.py."""

import pytest

from projectq import MainEngine
from projectq.cengines import InstructionFilter, AutoReplacer, DecompositionRuleSet
from projectq.backends import Simulator
from projectq.ops import All, BasicMathGate, ClassicalInstructionGate, Measure, X

import projectq.libs.math
from projectq.setups.decompositions import qft2crandhadamard, swap2cnot
from projectq.libs.math import AddConstant, AddConstantModN, MultiplyByConstantModN


def init(engine, quint, value):
    for i in range(len(quint)):
        if ((value >> i) & 1) == 1:
            X | quint[i]


def no_math_emulation(eng, cmd):
    if isinstance(cmd.gate, BasicMathGate):
        return False
    if isinstance(cmd.gate, ClassicalInstructionGate):
        return True
    try:
        return len(cmd.gate.matrix) == 2
    except AttributeError:
        return False


@pytest.fixture
def eng():
    return MainEngine(
        backend=Simulator(),
        engine_list=[AutoReplacer(rule_set), InstructionFilter(no_math_emulation)],
    )


rule_set = DecompositionRuleSet(modules=[projectq.libs.math, qft2crandhadamard, swap2cnot])


@pytest.mark.parametrize(
    'gate', (AddConstantModN(-1, 6), MultiplyByConstantModN(-1, 6), MultiplyByConstantModN(4, 4)), ids=str
)
def test_invalid(eng, gate):
    qureg = eng.allocate_qureg(4)
    init(eng, qureg, 4)

    with pytest.raises(ValueError):
        gate | qureg
        eng.flush()


def test_adder(eng):
    qureg = eng.allocate_qureg(4)
    init(eng, qureg, 4)

    AddConstant(3) | qureg

    assert 1.0 == pytest.approx(abs(eng.backend.cheat()[1][7]))

    init(eng, qureg, 7)  # reset
    init(eng, qureg, 2)

    # check for overflow -> should be 15+2 = 1 (mod 16)
    AddConstant(15) | qureg
    assert 1.0 == pytest.approx(abs(eng.backend.cheat()[1][1]))

    All(Measure) | qureg


def test_modadder(eng):
    qureg = eng.allocate_qureg(4)
    init(eng, qureg, 4)

    AddConstantModN(3, 6) | qureg

    assert 1.0 == pytest.approx(abs(eng.backend.cheat()[1][1]))

    init(eng, qureg, 1)  # reset
    init(eng, qureg, 7)

    AddConstantModN(10, 13) | qureg
    assert 1.0 == pytest.approx(abs(eng.backend.cheat()[1][4]))

    All(Measure) | qureg


def test_modmultiplier(eng):
    qureg = eng.allocate_qureg(4)
    init(eng, qureg, 4)

    MultiplyByConstantModN(3, 7) | qureg

    assert 1.0 == pytest.approx(abs(eng.backend.cheat()[1][5]))

    init(eng, qureg, 5)  # reset
    init(eng, qureg, 7)

    MultiplyByConstantModN(4, 13) | qureg
    assert 1.0 == pytest.approx(abs(eng.backend.cheat()[1][2]))

    All(Measure) | qureg



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

""" Back-end to run quantum programs using IonQ hardware."""
import random

from projectq.cengines import BasicEngine
from projectq.meta import LogicalQubitIDTag, get_control_count, has_negative_control
from projectq.ops import (
    Allocate,
    Barrier,
    DaggeredGate,
    Deallocate,
    FlushGate,
    HGate,
    Measure,
    R,
    Rx,
    Rxx,
    Ry,
    Ryy,
    Rz,
    Rzz,
    Sdag,
    SGate,
    SqrtXGate,
    SwapGate,
    Tdag,
    TGate,
    XGate,
    YGate,
    ZGate,
)
from projectq.types import WeakQubitRef

from . import _ionq_http_client as http_client
from ._ionq_exc import InvalidCommandError, MidCircuitMeasurementError

GATE_MAP = {
    XGate: 'x',
    YGate: 'y',
    ZGate: 'z',
    HGate: 'h',
    Rx: 'rx',
    Ry: 'ry',
    Rz: 'rz',
    SGate: 's',
    TGate: 't',
    SqrtXGate: 'v',
    Rxx: 'xx',
    Ryy: 'yy',
    Rzz: 'zz',
    SwapGate: 'swap',
}
SUPPORTED_GATES = tuple(GATE_MAP.keys())


def _rearrange_result(input_result, length):
    """Turn ``input_result`` from an integer into a bit-string.

    Args:
        input_result (int): An integer representation of qubit states.
        length (int): The total number of bits (for padding, if needed).

    Returns:
        str: A bit-string representation of ``input_result``.
    """
    bin_input = list(bin(input_result)[2:].rjust(length, '0'))
    return ''.join(bin_input)[::-1]


class IonQBackend(BasicEngine):  # pylint: disable=too-many-instance-attributes
    """Backend for building circuits and submitting them to the IonQ API."""

    def __init__(
        self,
        use_hardware=False,
        num_runs=100,
        verbose=False,
        token=None,
        device='ionq_simulator',
        num_retries=3000,
        interval=1,
        retrieve_execution=None,
    ):  # pylint: disable=too-many-arguments
        """Constructor for the IonQBackend.

        Args:
            use_hardware (bool, optional): Whether or not to use real IonQ
                hardware or just a simulator. If False, the ionq_simulator is
                used regardless of the value of ``device``. Defaults to False.
            num_runs (int, optional): Number of times to run circuits. Defaults to 100.
            verbose (bool, optional): If True, print statistics after job
                results have been collected. Defaults to False.
            token (str, optional): An IonQ API token. Defaults to None.
            device (str, optional): Device to run jobs on.
                Supported devices are ``'ionq_qpu'`` or ``'ionq_simulator'``.
                Defaults to ``'ionq_simulator'``.
            num_retries (int, optional): Number of times to retry fetching a
                job after it has been submitted. Defaults to 3000.
            interval (int, optional): Number of seconds to wait inbetween
                result fetch retries. Defaults to 1.
            retrieve_execution (str, optional): An IonQ API Job ID.
                If provided, a job with this ID will be fetched. Defaults to None.
        """
        BasicEngine.__init__(self)
        self.device = device if use_hardware else 'ionq_simulator'
        self._num_runs = num_runs
        self._verbose = verbose
        self._token = token
        self._num_retries = num_retries
        self._interval = interval
        self._circuit = []
        self._measured_ids = []
        self._probabilities = dict()
        self._retrieve_execution = retrieve_execution
        self._clear = True

    def is_available(self, cmd):
        """Test if this backend is available to process the provided command.

        Args:
            cmd (Command): A command to process.

        Returns:
            bool: If this backend can process the command.
        """
        gate = cmd.gate

        # Metagates.
        if gate in (Measure, Allocate, Deallocate, Barrier):
            return True

        if has_negative_control(cmd):
            return False

        # CNOT gates.
        # NOTE: IonQ supports up to 7 control qubits
        num_ctrl_qubits = get_control_count(cmd)
        if 0 < num_ctrl_qubits <= 7:
            return isinstance(gate, (XGate,))

        # Gates witout control bits.
        if num_ctrl_qubits == 0:
            supported = isinstance(gate, SUPPORTED_GATES)
            supported_transpose = gate in (Sdag, Tdag)
            return supported or supported_transpose
        return False

    def _reset(self):
        """Reset this backend.

        .. NOTE::

            This sets ``_clear = True``, which will trigger state cleanup
            on the next call to ``_store``.
        """

        # Lastly, reset internal state for measured IDs and circuit body.
        self._circuit = []
        self._clear = True

    def _store(self, cmd):
        """Interpret the ProjectQ command as a circuit instruction and store it.

        Args:
            cmd (Command): A command to process.

        Raises:
            InvalidCommandError: If the command can not be interpreted.
            MidCircuitMeasurementError: If this command would result in a
                mid-circuit qubit measurement.
        """
        if self._clear:
            self._measured_ids = []
            self._probabilities = dict()
            self._clear = False

        # No-op/Meta gates.
        # NOTE: self.main_engine.mapper takes care qubit allocation/mapping.
        gate = cmd.gate
        if gate in (Allocate, Deallocate, Barrier):
            return

        # Create a measurement.
        if gate == Measure:
            logical_id = cmd.qubits[0][0].id
            for tag in cmd.tags:
                if isinstance(tag, LogicalQubitIDTag):
                    logical_id = tag.logical_qubit_id
                    break
            # Add the qubit id
            self._measured_ids.append(logical_id)
            return

        # Process the Command's gate type:
        gate_type = type(gate)
        gate_name = GATE_MAP.get(gate_type)
        # Daggered gates get special treatment.
        if isinstance(gate, DaggeredGate):
            gate_name = GATE_MAP[type(gate._gate)] + 'i'  # pylint: disable=protected-access

        # Unable to determine a gate mapping here, so raise out.
        if gate_name is None:
            raise InvalidCommandError('Invalid command: ' + str(cmd))

        # Now make sure there are no existing measurements on qubits involved
        #   in this operation.
        targets = [qb.id for qureg in cmd.qubits for qb in qureg]
        controls = [qb.id for qb in cmd.control_qubits]
        if len(self._measured_ids) > 0:

            # Check any qubits we are trying to operate on.
            gate_qubits = set(targets) | set(controls)

            # If any of them have already been measured...
            already_measured = gate_qubits & set(self._measured_ids)

            # Boom!
            if len(already_measured) > 0:
                err = (
                    'Mid-circuit measurement is not supported. '
                    'The following qubits have already been measured: {}.'.format(list(already_measured))
                )
                raise MidCircuitMeasurementError(err)

        # Initialize the gate dict:
        gate_dict = {
            'gate': gate_name,
            'targets': targets,
        }

        # Check if we have a rotation
        if isinstance(gate, (R, Rx, Ry, Rz, Rxx, Ryy, Rzz)):
            gate_dict['rotation'] = gate.angle

        # Set controls
        if len(controls) > 0:
            gate_dict['controls'] = controls

        self._circuit.append(gate_dict)

    def get_probability(self, state, qureg):
        """Shortcut to get a specific state's probability.

        Args:
            state (str): A state in bit-string format.
            qureg (Qureg): A ProjectQ Qureg object.

        Returns:
            float: The probability for the provided state.
        """
        if len(state) != len(qureg):
            raise ValueError('Desired state and register must be the same length!')

        probs = self.get_probabilities(qureg)
        return probs[state]

    def get_probabilities(self, qureg):
        """Given the provided qubit register, determine the probability of
        each possible outcome.

        .. NOTE::

            This method should only be called *after* a circuit has been
            run and its results are available.

        Args:
            qureg (Qureg): A ProjectQ Qureg object.

        Returns:
            dict: A dict mapping of states -> probability.
        """
        if len(self._probabilities) == 0:
            raise RuntimeError("Please, run the circuit first!")

        probability_dict = {}
        for state in self._probabilities:
            mapped_state = ['0'] * len(qureg)
            for i, qubit in enumerate(qureg):
                try:
                    meas_idx = self._measured_ids.index(qubit.id)
                except ValueError:
                    continue
                mapped_state[i] = state[meas_idx]
            probability = self._probabilities[state]
            mapped_state = "".join(mapped_state)
            probability_dict[mapped_state] = probability_dict.get(mapped_state, 0) + probability
        return probability_dict

    def _run(self):  # pylint: disable=too-many-locals
        """Run the circuit this object has built during engine execution."""
        # Nothing to do with an empty circuit.
        if len(self._circuit) == 0:
            return

        if self._retrieve_execution is None:
            qubit_mapping = self.main_engine.mapper.current_mapping
            measured_ids = self._measured_ids[:]
            info = {
                'circuit': self._circuit,
                'nq': len(qubit_mapping.keys()),
                'shots': self._num_runs,
                'meas_mapped': [qubit_mapping[qubit_id] for qubit_id in measured_ids],
                'meas_qubit_ids': measured_ids,
            }
            res = http_client.send(
                info,
                device=self.device,
                token=self._token,
                num_retries=self._num_retries,
                interval=self._interval,
                verbose=self._verbose,
            )
            if res is None:
                raise RuntimeError('Failed to submit job to the server!')
        else:
            res = http_client.retrieve(
                device=self.device,
                token=self._token,
                jobid=self._retrieve_execution,
                num_retries=self._num_retries,
                interval=self._interval,
                verbose=self._verbose,
            )
            if res is None:
                raise RuntimeError("Failed to retrieve job with id: '{}'!".format(self._retrieve_execution))
            self._measured_ids = measured_ids = res['meas_qubit_ids']

        # Determine random outcome from probable states.
        random_outcome = random.random()
        p_sum = 0.0
        measured = ""
        star = ""
        num_measured = len(measured_ids)
        probable_outcomes = res['output_probs']
        states = probable_outcomes.keys()
        self._probabilities = {}
        for idx, state_int in enumerate(states):
            state = _rearrange_result(int(state_int), num_measured)
            probability = probable_outcomes[state_int]
            p_sum += probability
            if p_sum >= random_outcome and measured == "" or (idx == len(states) - 1):
                measured = state
                star = "*"
            self._probabilities[state] = probability
            if self._verbose and probability > 0:  # pragma: no cover
                print(state + " with p = " + str(probability) + star)

        # Register measurement results
        for idx, qubit_id in enumerate(measured_ids):
            result = int(measured[idx])
            qubit_ref = WeakQubitRef(self.main_engine, qubit_id)
            self.main_engine.set_measurement_result(qubit_ref, result)

    def receive(self, command_list):
        """Receive a command list from the ProjectQ engine pipeline.

        If a given command is a "flush" operation, the pending circuit will be
        submitted to IonQ's API for processing.

        Args:
            command_list (list[Command]): A list of ProjectQ Command objects.
        """
        for cmd in command_list:
            if not isinstance(cmd.gate, FlushGate):
                self._store(cmd)
            else:
                # After that, the circuit is ready to be submitted.
                try:
                    self._run()
                finally:
                    # Make sure we always reset engine state so as not to leave
                    #    anything dirty atexit.
                    self._reset()


__all__ = ['IonQBackend']



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"Tests for projectq.setups.decompositions.arb1qubit2rzandry.py."

from cmath import exp
import math

import numpy as np
import pytest

from projectq.backends import Simulator
from projectq.cengines import (
    AutoReplacer,
    DecompositionRuleSet,
    DummyEngine,
    InstructionFilter,
    MainEngine,
)
from projectq.ops import (
    BasicGate,
    ClassicalInstructionGate,
    MatrixGate,
    Measure,
    Ph,
    R,
    Rx,
    Ry,
    Rz,
    X,
)
from projectq.meta import Control

from . import arb1qubit2rzandry as arb1q


def test_recognize_correct_gates():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend)
    qubit = eng.allocate_qubit()
    Ph(0.1) | qubit
    R(0.2) | qubit
    Rx(0.3) | qubit
    X | qubit
    eng.flush(deallocate_qubits=True)
    # Don't test initial allocate and trailing deallocate and flush gate.
    for cmd in saving_backend.received_commands[1:-2]:
        assert arb1q._recognize_arb1qubit(cmd)


def test_recognize_incorrect_gates():
    saving_backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=saving_backend)
    qubit = eng.allocate_qubit()
    # Does not have matrix attribute:
    BasicGate() | qubit
    # Two qubit gate:
    two_qubit_gate = MatrixGate()
    two_qubit_gate.matrix = [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
    two_qubit_gate | qubit
    # Controlled single qubit gate:
    ctrl_qubit = eng.allocate_qubit()
    with Control(eng, ctrl_qubit):
        Rz(0.1) | qubit
    eng.flush(deallocate_qubits=True)
    for cmd in saving_backend.received_commands:
        assert not arb1q._recognize_arb1qubit(cmd)


def z_y_decomp_gates(eng, cmd):
    g = cmd.gate
    if isinstance(g, ClassicalInstructionGate):
        return True
    if len(cmd.control_qubits) == 0:
        if isinstance(cmd.gate, Ry) or isinstance(cmd.gate, Rz) or isinstance(cmd.gate, Ph):
            return True
    return False


def create_unitary_matrix(a, b, c, d):
    """
    Creates a unitary 2x2 matrix given parameters.

    Any unitary 2x2 matrix can be parametrized by:
    U = exp(ia) [[exp(j*b) * cos(d), exp(j*c) * sin(d)],
                 [-exp(-j*c) * sin(d), exp(-j*b) * cos(d)]]
    with 0 <= d <= pi/2 and 0 <= a,b,c < 2pi. If a==0, then
    det(U) == 1 and hence U is element of SU(2).

    Args:
        a,b,c (float): parameters 0 <= a,b,c < 2pi
        d (float): parameter 0 <= d <= pi/2

    Returns:
        2x2 matrix as nested lists
    """
    ph = exp(1j * a)  # global phase
    return [
        [ph * exp(1j * b) * math.cos(d), ph * exp(1j * c) * math.sin(d)],
        [ph * -exp(-1j * c) * math.sin(d), ph * exp(-1j * b) * math.cos(d)],
    ]


def create_test_matrices():
    params = [
        (0.2, 0.3, 0.5, math.pi * 0.4),
        (1e-14, 0.3, 0.5, 0),
        (0.4, 0.0, math.pi * 2, 0.7),
        (0.0, 0.2, math.pi * 1.2, 1.5),  # element of SU(2)
        (0.4, 0.0, math.pi * 1.3, 0.8),
        (0.4, 4.1, math.pi * 1.3, 0),
        (5.1, 1.2, math.pi * 1.5, math.pi / 2.0),
        (1e-13, 1.2, math.pi * 3.7, math.pi / 2.0),
        (0, math.pi / 2.0, 0, 0),
        (math.pi / 2.0, -math.pi / 2.0, 0, 0),
        (math.pi / 2.0, math.pi / 2.0, 0.1, 0.4),
        (math.pi * 1.5, math.pi / 2.0, 0, 0.4),
    ]
    matrices = []
    for a, b, c, d in params:
        matrices.append(create_unitary_matrix(a, b, c, d))
    return matrices


@pytest.mark.parametrize("gate_matrix", create_test_matrices())
def test_decomposition(gate_matrix):
    for basis_state in ([1, 0], [0, 1]):
        # Create single qubit gate with gate_matrix
        test_gate = MatrixGate()
        test_gate.matrix = np.matrix(gate_matrix)

        correct_dummy_eng = DummyEngine(save_commands=True)
        correct_eng = MainEngine(backend=Simulator(), engine_list=[correct_dummy_eng])

        rule_set = DecompositionRuleSet(modules=[arb1q])
        test_dummy_eng = DummyEngine(save_commands=True)
        test_eng = MainEngine(
            backend=Simulator(),
            engine_list=[
                AutoReplacer(rule_set),
                InstructionFilter(z_y_decomp_gates),
                test_dummy_eng,
            ],
        )

        correct_qb = correct_eng.allocate_qubit()
        correct_eng.flush()
        test_qb = test_eng.allocate_qubit()
        test_eng.flush()

        correct_eng.backend.set_wavefunction(basis_state, correct_qb)
        test_eng.backend.set_wavefunction(basis_state, test_qb)

        test_gate | test_qb
        test_gate | correct_qb

        test_eng.flush()
        correct_eng.flush()

        assert correct_dummy_eng.received_commands[2].gate == test_gate
        assert test_dummy_eng.received_commands[2].gate != test_gate

        for fstate in ['0', '1']:
            test = test_eng.backend.get_amplitude(fstate, test_qb)
            correct = correct_eng.backend.get_amplitude(fstate, correct_qb)
            assert correct == pytest.approx(test, rel=1e-12, abs=1e-12)

        Measure | test_qb
        Measure | correct_qb


@pytest.mark.parametrize("gate_matrix", [[[2, 0], [0, 4]], [[0, 2], [4, 0]], [[1, 2], [4, 0]]])
def test_decomposition_errors(gate_matrix):
    test_gate = MatrixGate()
    test_gate.matrix = np.matrix(gate_matrix)
    rule_set = DecompositionRuleSet(modules=[arb1q])
    eng = MainEngine(
        backend=DummyEngine(),
        engine_list=[AutoReplacer(rule_set), InstructionFilter(z_y_decomp_gates)],
    )
    qb = eng.allocate_qubit()
    with pytest.raises(Exception):
        test_gate | qb



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pytest

from projectq import MainEngine
from projectq.ops import (
    All,
    BasicMathGate,
    C,
    Measure,
    NOT,
    X,
    Y,
)
from projectq.cengines import (
    AutoReplacer,
    BasicMapperEngine,
    DecompositionRuleSet,
    DummyEngine,
)
from ._simulator_test import mapper  # noqa: F401
from projectq.types import WeakQubitRef

from ._classical_simulator import ClassicalSimulator


def test_simulator_read_write(mapper):  # noqa: F811
    engine_list = []
    if mapper is not None:
        engine_list.append(mapper)
    sim = ClassicalSimulator()
    eng = MainEngine(sim, engine_list)
    a = eng.allocate_qureg(32)
    b = eng.allocate_qureg(32)

    assert sim.read_register(a) == 0
    assert sim.read_register(b) == 0
    assert sim.read_bit(a[0]) == 0
    assert sim.read_bit(b[0]) == 0

    sim.write_register(a, 123)
    sim.write_register(b, 456)
    assert sim.read_register(a) == 123
    assert sim.read_register(b) == 456
    assert sim.read_bit(a[0]) == 1
    assert sim.read_bit(b[0]) == 0

    sim.write_bit(b[0], 1)
    assert sim.read_register(a) == 123
    assert sim.read_register(b) == 457
    assert sim.read_bit(a[0]) == 1
    assert sim.read_bit(b[0]) == 1


def test_simulator_triangle_increment_cycle(mapper):  # noqa: F811
    engine_list = []
    if mapper is not None:
        engine_list.append(mapper)
    sim = ClassicalSimulator()
    eng = MainEngine(sim, engine_list)

    a = eng.allocate_qureg(6)
    for t in range(1 << 6):
        assert sim.read_register(a) == t
        for i in range(6)[::-1]:
            C(X, i) | (a[:i], a[i])
    assert sim.read_register(a) == 0


def test_simulator_bit_repositioning(mapper):  # noqa: F811
    engine_list = []
    if mapper is not None:
        engine_list.append(mapper)
    sim = ClassicalSimulator()
    eng = MainEngine(sim, engine_list)
    a = eng.allocate_qureg(4)
    b = eng.allocate_qureg(5)
    c = eng.allocate_qureg(6)
    sim.write_register(a, 9)
    sim.write_register(b, 17)
    sim.write_register(c, 33)
    for q in b:
        eng.deallocate_qubit(q)
        # Make sure that the qubit are marked as deleted
        assert q.id == -1
    assert sim.read_register(a) == 9
    assert sim.read_register(c) == 33


def test_simulator_arithmetic(mapper):  # noqa: F811
    class Offset(BasicMathGate):
        def __init__(self, amount):
            BasicMathGate.__init__(self, lambda x: (x + amount,))

    class Sub(BasicMathGate):
        def __init__(self):
            BasicMathGate.__init__(self, lambda x, y: (x, y - x))

    engine_list = []
    if mapper is not None:
        engine_list.append(mapper)
    sim = ClassicalSimulator()
    eng = MainEngine(sim, engine_list)
    a = eng.allocate_qureg(4)
    b = eng.allocate_qureg(5)
    sim.write_register(a, 9)
    sim.write_register(b, 17)

    Offset(2) | a
    assert sim.read_register(a) == 11
    assert sim.read_register(b) == 17

    Offset(3) | b
    assert sim.read_register(a) == 11
    assert sim.read_register(b) == 20

    Offset(32 + 5) | b
    assert sim.read_register(a) == 11
    assert sim.read_register(b) == 25

    Sub() | (a, b)
    assert sim.read_register(a) == 11
    assert sim.read_register(b) == 14

    Sub() | (a, b)
    Sub() | (a, b)
    assert sim.read_register(a) == 11
    assert sim.read_register(b) == 24

    # also test via measurement:
    All(Measure) | a + b
    eng.flush()

    for i in range(len(a)):
        assert int(a[i]) == ((11 >> i) & 1)
    for i in range(len(b)):
        assert int(b[i]) == ((24 >> i) & 1)


def test_write_register_value_error_exception(mapper):  # noqa: F811
    engine_list = []
    if mapper is not None:
        engine_list.append(mapper)
    sim = ClassicalSimulator()
    eng = MainEngine(sim, engine_list)
    a = eng.allocate_qureg(3)
    with pytest.raises(ValueError):
        sim.write_register(a, -2)
    sim.write_register(a, 7)
    with pytest.raises(ValueError):
        sim.write_register(a, 8)


def test_x_gate_invalid():
    sim = ClassicalSimulator()
    eng = MainEngine(sim, [AutoReplacer(DecompositionRuleSet())])
    a = eng.allocate_qureg(2)

    with pytest.raises(ValueError):
        X | a


def test_available_gates():
    sim = ClassicalSimulator()
    eng = MainEngine(sim, [AutoReplacer(DecompositionRuleSet())])
    a = eng.allocate_qubit()
    X | a
    NOT | a
    Measure | a
    eng.flush()


def test_gates_are_forwarded_to_next_engine():
    sim = ClassicalSimulator()
    saving_eng = DummyEngine(save_commands=True)
    eng = MainEngine(saving_eng, [sim])
    a = eng.allocate_qubit()
    X | a
    a[0].__del__()
    assert len(saving_eng.received_commands) == 3


def test_wrong_gate():
    sim = ClassicalSimulator()
    eng = MainEngine(sim, [])
    a = eng.allocate_qubit()
    with pytest.raises(ValueError):
        Y | a


def test_runtime_error():
    sim = ClassicalSimulator()
    mapper = BasicMapperEngine()  # noqa: F811
    mapper.current_mapping = {}
    eng = MainEngine(sim, [mapper])
    with pytest.raises(RuntimeError):
        eng.backend.read_bit(WeakQubitRef(None, 1))



# -*- coding: utf-8 -*-
# pylint: skip-file

from projectq.cengines import MainEngine
from projectq.ops import All, H, X, Measure
from projectq.meta import Compute, Uncompute, Dagger
from projectq.libs.revkit import PhaseOracle, PermutationOracle

import revkit


# phase function
def f(a, b, c, d, e, f):
    return (a and b) ^ (c and d) ^ (e and f)


# permutation
pi = [0, 2, 3, 5, 7, 1, 4, 6]

eng = MainEngine()
qubits = eng.allocate_qureg(6)
x = qubits[::2]  # qubits on odd lines
y = qubits[1::2]  # qubits on even lines

# circuit
with Compute(eng):
    All(H) | qubits
    All(X) | [x[0], x[1]]
    PermutationOracle(pi) | y
PhaseOracle(f) | qubits
Uncompute(eng)

with Compute(eng):
    with Dagger(eng):
        PermutationOracle(pi, synth=revkit.dbs) | x
PhaseOracle(f) | qubits
Uncompute(eng)

All(H) | qubits

All(Measure) | qubits

# measurement result
print("Shift is {}".format(sum(int(q) << i for i, q in enumerate(qubits))))



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""ProjectQ module containing all basic gates (operations)"""

from ._basics import (
    NotMergeable,
    NotInvertible,
    BasicGate,
    MatrixGate,
    SelfInverseGate,
    BasicRotationGate,
    ClassicalInstructionGate,
    FastForwardingGate,
    BasicMathGate,
    BasicPhaseGate,
)
from ._command import apply_command, Command, CtrlAll, IncompatibleControlState
from ._metagates import DaggeredGate, get_inverse, is_identity, ControlledGate, C, Tensor, All
from ._gates import *
from ._qftgate import QFT, QFTGate
from ._qubit_operator import QubitOperator
from ._shortcuts import *
from ._time_evolution import TimeEvolution
from ._uniformly_controlled_rotation import UniformlyControlledRy, UniformlyControlledRz
from ._state_prep import StatePreparation
from ._qpegate import QPE
from ._qaagate import QAA



# -*- coding: utf-8 -*-
# pylint: skip-file

from projectq.ops import CNOT, H, Measure, Rz, X, Z
from projectq import MainEngine
from projectq.meta import Dagger, Control


def create_bell_pair(eng):
    r"""
    Returns a Bell-pair (two qubits in state :math:`|A\rangle \otimes |B
    \rangle = \frac 1{\sqrt 2} \left( |0\rangle\otimes|0\rangle + |1\rangle
    \otimes|1\rangle \right)`).

    Args:
        eng (MainEngine): MainEngine from which to allocate the qubits.

    Returns:
        bell_pair (tuple<Qubits>): The Bell-pair.
    """
    b1 = eng.allocate_qubit()
    b2 = eng.allocate_qubit()

    H | b1
    CNOT | (b1, b2)

    return b1, b2


def run_teleport(eng, state_creation_function, verbose=False):
    """
    Runs quantum teleportation on the provided main compiler engine.

    Creates a state from |0> using the state_creation_function, teleports this
    state to Bob who then tries to uncompute his qubit using the inverse of
    the state_creation_function. If successful, deleting the qubit won't raise
    an error in the underlying Simulator back-end (else it will).

    Args:
        eng (MainEngine): Main compiler engine to run the circuit on.
        state_creation_function (function): Function which accepts the main
            engine and a qubit in state |0>, which it then transforms to the
            state that Alice would like to send to Bob.
        verbose (bool): If True, info messages will be printed.

    """
    # make a Bell-pair
    b1, b2 = create_bell_pair(eng)

    # Alice creates a nice state to send
    psi = eng.allocate_qubit()
    if verbose:
        print("Alice is creating her state from scratch, i.e., |0>.")
    state_creation_function(eng, psi)

    # entangle it with Alice's b1
    CNOT | (psi, b1)
    if verbose:
        print("Alice entangled her qubit with her share of the Bell-pair.")

    # measure two values (once in Hadamard basis) and send the bits to Bob
    H | psi
    Measure | psi
    Measure | b1
    msg_to_bob = [int(psi), int(b1)]
    if verbose:
        print("Alice is sending the message {} to Bob.".format(msg_to_bob))

    # Bob may have to apply up to two operation depending on the message sent
    # by Alice:
    with Control(eng, b1):
        X | b2
    with Control(eng, psi):
        Z | b2

    # try to uncompute the psi state
    if verbose:
        print("Bob is trying to uncompute the state.")
    with Dagger(eng):
        state_creation_function(eng, b2)

    # check whether the uncompute was successful. The simulator only allows to
    # delete qubits which are in a computational basis state.
    del b2
    eng.flush()

    if verbose:
        print("Bob successfully arrived at |0>")


if __name__ == "__main__":
    # create a main compiler engine with a simulator backend:
    eng = MainEngine()

    # define our state-creation routine, which transforms a |0> to the state
    # we would like to send. Bob can then try to uncompute it and, if he
    # arrives back at |0>, we know that the teleportation worked.
    def create_state(eng, qb):
        H | qb
        Rz(1.21) | qb

    # run the teleport and then, let Bob try to uncompute his qubit:
    run_teleport(eng, create_state, verbose=True)



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Tests for decompositions rules (using the Simulator).
"""

import pytest

from projectq.cengines import (
    MainEngine,
    InstructionFilter,
    AutoReplacer,
    DummyEngine,
    DecompositionRuleSet,
)
from projectq.backends import Simulator
from projectq.ops import (
    All,
    ClassicalInstructionGate,
    CRz,
    Entangle,
    H,
    Measure,
    Ph,
    R,
    Rz,
    T,
    Tdag,
    Toffoli,
    X,
)
from projectq.meta import Control
from projectq.setups.decompositions import (
    crz2cxandrz,
    entangle,
    globalphase,
    ph2r,
    r2rzandph,
    toffoli2cnotandtgate,
)


def low_level_gates(eng, cmd):
    g = cmd.gate
    if isinstance(g, ClassicalInstructionGate):
        return True
    if len(cmd.control_qubits) == 0:
        if g == T or g == Tdag or g == H or isinstance(g, Rz) or isinstance(g, Ph):
            return True
    else:
        if len(cmd.control_qubits) == 1 and cmd.gate == X:
            return True
    return False


def test_entangle():
    rule_set = DecompositionRuleSet(modules=[entangle])
    sim = Simulator()
    eng = MainEngine(sim, [AutoReplacer(rule_set), InstructionFilter(low_level_gates)])
    qureg = eng.allocate_qureg(4)
    Entangle | qureg

    assert 0.5 == pytest.approx(abs(sim.cheat()[1][0]) ** 2)
    assert 0.5 == pytest.approx(abs(sim.cheat()[1][-1]) ** 2)

    All(Measure) | qureg


def low_level_gates_noglobalphase(eng, cmd):
    return low_level_gates(eng, cmd) and not isinstance(cmd.gate, Ph) and not isinstance(cmd.gate, R)


def test_globalphase():
    rule_set = DecompositionRuleSet(modules=[globalphase, r2rzandph])
    dummy = DummyEngine(save_commands=True)
    eng = MainEngine(
        dummy,
        [AutoReplacer(rule_set), InstructionFilter(low_level_gates_noglobalphase)],
    )

    qubit = eng.allocate_qubit()
    R(1.2) | qubit

    rz_count = 0
    for cmd in dummy.received_commands:
        assert not isinstance(cmd.gate, R)
        if isinstance(cmd.gate, Rz):
            rz_count += 1
            assert cmd.gate == Rz(1.2)

    assert rz_count == 1


def run_circuit(eng):
    qureg = eng.allocate_qureg(4)
    All(H) | qureg
    CRz(3.0) | (qureg[0], qureg[1])
    Toffoli | (qureg[1], qureg[2], qureg[3])

    with Control(eng, qureg[0:2]):
        Ph(1.43) | qureg[2]
    return qureg


def test_gate_decompositions():
    sim = Simulator()
    eng = MainEngine(sim, [])
    rule_set = DecompositionRuleSet(modules=[r2rzandph, crz2cxandrz, toffoli2cnotandtgate, ph2r])

    qureg = run_circuit(eng)

    sim2 = Simulator()
    eng_lowlevel = MainEngine(sim2, [AutoReplacer(rule_set), InstructionFilter(low_level_gates)])
    qureg2 = run_circuit(eng_lowlevel)

    for i in range(len(sim.cheat()[1])):
        assert sim.cheat()[1][i] == pytest.approx(sim2.cheat()[1][i])

    All(Measure) | qureg
    All(Measure) | qureg2



# -*- coding: utf-8 -*-
#   Copyright 2020 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Math gates for ProjectQ"""

from projectq.ops import BasicMathGate


class AddConstant(BasicMathGate):
    """
    Add a constant to a quantum number represented by a quantum register,
    stored from low- to high-bit.

    Example:
        .. code-block:: python

            qunum = eng.allocate_qureg(5) # 5-qubit number
            X | qunum[1] # qunum is now equal to 2
            AddConstant(3) | qunum # qunum is now equal to 5

    Important: if you run with conditional and carry, carry needs to
    be a quantum register for the compiler/decomposition to work.
    """

    def __init__(self, a):  # pylint: disable=invalid-name
        """
        Initializes the gate to the number to add.

        Args:
            a (int): Number to add to a quantum register.

        It also initializes its base class, BasicMathGate, with the
        corresponding function, so it can be emulated efficiently.
        """
        BasicMathGate.__init__(self, lambda x: ((x + a),))
        self.a = a  # pylint: disable=invalid-name

    def get_inverse(self):
        """
        Return the inverse gate (subtraction of the same constant).
        """
        return SubConstant(self.a)

    def __str__(self):
        return "AddConstant({})".format(self.a)

    def __eq__(self, other):
        return isinstance(other, AddConstant) and self.a == other.a

    def __hash__(self):
        return hash(str(self))

    def __ne__(self, other):
        return not self.__eq__(other)


def SubConstant(a):  # pylint: disable=invalid-name
    """
    Subtract a constant from a quantum number represented by a quantum
    register, stored from low- to high-bit.

    Args:
        a (int): Constant to subtract

    Example:
        .. code-block:: python

            qunum = eng.allocate_qureg(5) # 5-qubit number
            X | qunum[2] # qunum is now equal to 4
            SubConstant(3) | qunum # qunum is now equal to 1
    """
    return AddConstant(-a)


class AddConstantModN(BasicMathGate):
    """
    Add a constant to a quantum number represented by a quantum register
    modulo N.

    The number is stored from low- to high-bit, i.e., qunum[0] is the LSB.

    Example:
        .. code-block:: python

            qunum = eng.allocate_qureg(5) # 5-qubit number
            X | qunum[1] # qunum is now equal to 2
            AddConstantModN(3, 4) | qunum # qunum is now equal to 1

    .. note::

      Pre-conditions:

      * c < N
      * c >= 0
      * The value stored in the quantum register must be lower than N
    """

    def __init__(self, a, N):
        """
        Initializes the gate to the number to add modulo N.

        Args:
            a (int): Number to add to a quantum register (0 <= a < N).
            N (int): Number modulo which the addition is carried out.

        It also initializes its base class, BasicMathGate, with the
        corresponding function, so it can be emulated efficiently.
        """
        BasicMathGate.__init__(self, lambda x: ((x + a) % N,))
        self.a = a  # pylint: disable=invalid-name
        self.N = N

    def __str__(self):
        return "AddConstantModN({}, {})".format(self.a, self.N)

    def get_inverse(self):
        """
        Return the inverse gate (subtraction of the same number a modulo the
        same number N).
        """
        return SubConstantModN(self.a, self.N)

    def __eq__(self, other):
        return isinstance(other, AddConstantModN) and self.a == other.a and self.N == other.N

    def __hash__(self):
        return hash(str(self))

    def __ne__(self, other):
        return not self.__eq__(other)


def SubConstantModN(a, N):  # pylint: disable=invalid-name
    """
    Subtract a constant from a quantum number represented by a quantum
    register modulo N.

    The number is stored from low- to high-bit, i.e., qunum[0] is the LSB.

    Args:
        a (int): Constant to add
        N (int): Constant modulo which the addition of a should be carried
            out.

    Example:
        .. code-block:: python

            qunum = eng.allocate_qureg(3) # 3-qubit number
            X | qunum[1] # qunum is now equal to 2
            SubConstantModN(4,5) | qunum # qunum is now -2 = 6 = 1 (mod 5)

    .. note::

      Pre-conditions:

      * c < N
      * c >= 0
      * The value stored in the quantum register must be lower than N
    """
    return AddConstantModN(N - a, N)


class MultiplyByConstantModN(BasicMathGate):
    """
    Multiply a quantum number represented by a quantum register by a constant
    modulo N.

    The number is stored from low- to high-bit, i.e., qunum[0] is the LSB.

    Example:
        .. code-block:: python

            qunum = eng.allocate_qureg(5) # 5-qubit number
            X | qunum[2] # qunum is now equal to 4
            MultiplyByConstantModN(3,5) | qunum # qunum is now 2.

    .. note::

      Pre-conditions:

      * c < N
      * c >= 0
      * gcd(c, N) == 1
      * The value stored in the quantum register must be lower than N
    """

    def __init__(self, a, N):  # pylint: disable=invalid-name
        """
        Initializes the gate to the number to multiply with modulo N.

        Args:
            a (int): Number by which to multiply a quantum register
                (0 <= a < N).
            N (int): Number modulo which the multiplication is carried out.

        It also initializes its base class, BasicMathGate, with the
        corresponding function, so it can be emulated efficiently.
        """
        BasicMathGate.__init__(self, lambda x: ((a * x) % N,))
        self.a = a  # pylint: disable=invalid-name
        self.N = N

    def __str__(self):
        return "MultiplyByConstantModN({}, {})".format(self.a, self.N)

    def __eq__(self, other):
        return isinstance(other, MultiplyByConstantModN) and self.a == other.a and self.N == other.N

    def __hash__(self):
        return hash(str(self))

    def __ne__(self, other):
        return not self.__eq__(other)


class AddQuantumGate(BasicMathGate):
    """
    Adds up two quantum numbers represented by quantum registers.
    The numbers are stored from low- to high-bit, i.e., qunum[0] is the LSB.

    Example:
        .. code-block:: python

            qunum_a = eng.allocate_qureg(5) # 5-qubit number
            qunum_b = eng.allocate_qureg(5) # 5-qubit number
            carry_bit = eng.allocate_qubit()

            X | qunum_a[2] #qunum_a is now equal to 4
            X | qunum_b[3] #qunum_b is now equal to 8
            AddQuantum | (qunum_a, qunum_b, carry)
            # qunum_a remains 4, qunum_b is now 12 and carry_bit is 0
    """

    def __init__(self):
        BasicMathGate.__init__(self, None)

    def __str__(self):
        return "AddQuantum"

    def __eq__(self, other):
        return isinstance(other, AddQuantumGate)

    def __hash__(self):
        return hash(str(self))

    def __ne__(self, other):
        return not self.__eq__(other)

    def get_math_function(self, qubits):
        n_qubits = len(qubits[0])

        def math_fun(a):  # pylint: disable=invalid-name
            a[1] = a[0] + a[1]
            if len(bin(a[1])[2:]) > n_qubits:
                a[1] = a[1] % (2 ** n_qubits)

                if len(a) == 3:
                    # Flip the last bit of the carry register
                    a[2] ^= 1
            return a

        return math_fun

    def get_inverse(self):
        """
        Return the inverse gate (subtraction of the same number a modulo the
        same number N).
        """
        return _InverseAddQuantumGate()


AddQuantum = AddQuantumGate()


class _InverseAddQuantumGate(BasicMathGate):
    """
    Internal gate glass to support emulation for inverse
    addition.
    """

    def __init__(self):
        BasicMathGate.__init__(self, None)

    def __str__(self):
        return "_InverseAddQuantum"

    def get_math_function(self, qubits):
        def math_fun(a):  # pylint: disable=invalid-name
            if len(a) == 3:
                # Flip the last bit of the carry register
                a[2] ^= 1

            a[1] -= a[0]
            return a

        return math_fun


class SubtractQuantumGate(BasicMathGate):
    """
    Subtract one quantum number represented by a quantum register from
    another quantum number represented by a quantum register.

    Example:
    .. code-block:: python

            qunum_a = eng.allocate_qureg(5) # 5-qubit number
            qunum_b = eng.allocate_qureg(5) # 5-qubit number
            X | qunum_a[2] #qunum_a is now equal to 4
            X | qunum_b[3] #qunum_b is now equal to 8
            SubtractQuantum | (qunum_a, qunum_b)
            # qunum_a remains 4, qunum_b is now 4

    """

    def __init__(self):
        """
        Initializes the gate to its base class, BasicMathGate, with the corresponding function, so it can be emulated
        efficiently.
        """

        def subtract(a, b):  # pylint: disable=invalid-name
            return (a, b - a)

        BasicMathGate.__init__(self, subtract)

    def __str__(self):
        return "SubtractQuantum"

    def __eq__(self, other):
        return isinstance(other, SubtractQuantumGate)

    def __hash__(self):
        return hash(str(self))

    def __ne__(self, other):
        return not self.__eq__(other)

    def get_inverse(self):
        """
        Return the inverse gate (subtraction of the same number a modulo the same number N).
        """
        return AddQuantum


SubtractQuantum = SubtractQuantumGate()


class ComparatorQuantumGate(BasicMathGate):
    """
    Flips a compare qubit if the binary value of first imput is higher than the second input.
    The numbers are stored from low- to high-bit, i.e., qunum[0] is the LSB.
    Example:
        .. code-block:: python

            qunum_a = eng.allocate_qureg(5) # 5-qubit number
            qunum_b = eng.allocate_qureg(5) # 5-qubit number
            compare_bit = eng.allocate_qubit()
            X | qunum_a[4] #qunum_a is now equal to 16
            X | qunum_b[3] #qunum_b is now equal to 8
            ComparatorQuantum | (qunum_a, qunum_b, compare_bit)
            # qunum_a and qunum_b remain 16 and 8, qunum_b is now 12 and
            compare bit is now 1

    """

    def __init__(self):
        """
        Initializes the gate and its base class, BasicMathGate, with the corresponding function, so it can be emulated
        efficiently.
        """

        def compare(a, b, c):  # pylint: disable=invalid-name
            # pylint: disable=invalid-name
            if b < a:
                if c == 0:
                    c = 1
                else:
                    c = 0
            return (a, b, c)

        BasicMathGate.__init__(self, compare)

    def __str__(self):
        return "Comparator"

    def __eq__(self, other):
        return isinstance(other, ComparatorQuantumGate)

    def __hash__(self):
        return hash(str(self))

    def __ne__(self, other):
        return not self.__eq__(other)

    def get_inverse(self):
        """
        Return the inverse gate
        """
        return AddQuantum


ComparatorQuantum = ComparatorQuantumGate()


class DivideQuantumGate(BasicMathGate):
    """
    Divides one quantum number from another. Takes three inputs which should be quantum registers of equal size; a
    dividend, a remainder and a divisor. The remainder should be in the state |0...0> and the dividend should be
    bigger than the divisor.The gate returns (in this order): the remainder, the quotient and the divisor.

    The numbers are stored from low- to high-bit, i.e., qunum[0] is the LSB.

    Example:
    .. code-block:: python

            qunum_a = eng.allocate_qureg(5) # 5-qubit number
            qunum_b = eng.allocate_qureg(5) # 5-qubit number
            qunum_c = eng.allocate_qureg(5) # 5-qubit number

            All(X) | [qunum_a[0],qunum_a[3]] #qunum_a is now equal to 9
            X | qunum_c[2] #qunum_c is now equal to 4

            DivideQuantum | (qunum_a, qunum_b,qunum_c)
            # qunum_a is now equal to 1 (remainder), qunum_b is now
            # equal to 2 (quotient) and qunum_c remains 4 (divisor)

            |dividend>|remainder>|divisor> -> |remainder>|quotient>|divisor>
    """

    def __init__(self):
        """
        Initializes the gate and its base class, BasicMathGate, with the corresponding function, so it can be emulated
        efficiently.
        """

        def division(dividend, remainder, divisor):
            if divisor == 0 or divisor > dividend:
                return (remainder, dividend, divisor)

            quotient = remainder + dividend // divisor
            return ((dividend - (quotient * divisor)), quotient, divisor)

        BasicMathGate.__init__(self, division)

    def get_inverse(self):
        return _InverseDivideQuantumGate()

    def __str__(self):
        return "DivideQuantum"

    def __eq__(self, other):
        return isinstance(other, DivideQuantumGate)

    def __hash__(self):
        return hash(str(self))

    def __ne__(self, other):
        return not self.__eq__(other)


DivideQuantum = DivideQuantumGate()


class _InverseDivideQuantumGate(BasicMathGate):
    """
    Internal gate glass to support emulation for inverse division.
    """

    def __init__(self):
        def inverse_division(remainder, quotient, divisor):
            if divisor == 0:
                return (quotient, remainder, divisor)

            dividend = remainder + quotient * divisor
            remainder = 0
            return (dividend, remainder, divisor)

        BasicMathGate.__init__(self, inverse_division)

    def __str__(self):
        return "_InverseDivideQuantum"


class MultiplyQuantumGate(BasicMathGate):
    """
    Multiplies two quantum numbers represented by a quantum registers.  Requires three quantum registers as inputs,
    the first two are the numbers to be multiplied and should have the same size (n qubits). The third register will
    hold the product and should be of size 2n+1.  The numbers are stored from low- to high-bit, i.e., qunum[0] is the
    LSB.

    Example:
        .. code-block:: python

        qunum_a = eng.allocate_qureg(4)
        qunum_b = eng.allocate_qureg(4)
        qunum_c = eng.allocate_qureg(9)
        X | qunum_a[2] # qunum_a is now 4
        X | qunum_b[3] # qunum_b is now 8
        MultiplyQuantum() | (qunum_a, qunum_b, qunum_c)
        # qunum_a remains 4 and qunum_b remains 8, qunum_c is now equal to 32
    """

    def __init__(self):
        """
        Initializes the gate and its base class, BasicMathGate, with the corresponding function, so it can be emulated
        efficiently.
        """

        def multiply(a, b, c):  # pylint: disable=invalid-name
            return (a, b, c + a * b)

        BasicMathGate.__init__(self, multiply)

    def __str__(self):
        return "MultiplyQuantum"

    def __eq__(self, other):
        return isinstance(other, MultiplyQuantumGate)

    def __hash__(self):
        return hash(str(self))

    def __ne__(self, other):
        return not self.__eq__(other)

    def get_inverse(self):
        return _InverseMultiplyQuantumGate()


MultiplyQuantum = MultiplyQuantumGate()


class _InverseMultiplyQuantumGate(BasicMathGate):
    """
    Internal gate glass to support emulation for inverse multiplication.
    """

    def __init__(self):
        def inverse_multiplication(a, b, c):  # pylint: disable=invalid-name
            return (a, b, c - a * b)

        BasicMathGate.__init__(self, inverse_multiplication)

    def __str__(self):
        return "_InverseMultiplyQuantum"



# -*- coding: utf-8 -*-
#   Copyright 2020 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pytest

from projectq import MainEngine
from projectq.cengines import InstructionFilter, AutoReplacer, DecompositionRuleSet
from projectq.backends import Simulator
from projectq.ops import All, BasicMathGate, ClassicalInstructionGate, Measure, X

from projectq.setups.decompositions import swap2cnot

import projectq.libs.math
from projectq.libs.math import (
    AddQuantum,
    SubtractQuantum,
    ComparatorQuantum,
    DivideQuantum,
    MultiplyQuantum,
)

from projectq.meta import Control, Compute, Uncompute, Dagger


def print_all_probabilities(eng, qureg):
    i = 0
    y = len(qureg)
    while i < (2 ** y):
        qubit_list = [int(x) for x in list(('{0:0b}'.format(i)).zfill(y))]
        qubit_list = qubit_list[::-1]
        prob = eng.backend.get_probability(qubit_list, qureg)
        if prob != 0.0:
            print(prob, qubit_list, i)
        i += 1


def init(engine, quint, value):
    for i in range(len(quint)):
        if ((value >> i) & 1) == 1:
            X | quint[i]


def no_math_emulation(eng, cmd):
    if isinstance(cmd.gate, BasicMathGate):
        return False
    if isinstance(cmd.gate, ClassicalInstructionGate):
        return True
    try:
        return len(cmd.gate.matrix) == 2
    except AttributeError:
        return False


rule_set = DecompositionRuleSet(modules=[projectq.libs.math, swap2cnot])


@pytest.fixture
def eng():
    return MainEngine(
        backend=Simulator(),
        engine_list=[AutoReplacer(rule_set), InstructionFilter(no_math_emulation)],
    )


@pytest.mark.parametrize('carry', (False, True))
@pytest.mark.parametrize('inverse', (False, True))
@pytest.mark.parametrize('qubit_idx', (0, 1, 2))
def test_quantumadder_size_mismatch(eng, qubit_idx, inverse, carry):
    qureg_a = eng.allocate_qureg(4 if qubit_idx != 0 else 3)
    qureg_b = eng.allocate_qureg(4 if qubit_idx != 1 else 3)
    qureg_c = eng.allocate_qureg(1 if qubit_idx != 2 else 2)

    if carry and inverse:
        pytest.skip('Inverse addition with carry not supported')
    elif not carry and qubit_idx == 2:
        pytest.skip('Invalid test parameter combination')

    with pytest.raises(ValueError):
        if inverse:
            with Dagger(eng):
                AddQuantum | (qureg_a, qureg_b, qureg_c if carry else [])
        else:
            AddQuantum | (qureg_a, qureg_b, qureg_c if carry else [])


@pytest.mark.parametrize('qubit_idx', (0, 1, 2))
def test_quantum_conditional_adder_size_mismatch(eng, qubit_idx):
    qureg_a = eng.allocate_qureg(4 if qubit_idx != 0 else 3)
    qureg_b = eng.allocate_qureg(4 if qubit_idx != 1 else 3)
    control = eng.allocate_qureg(1 if qubit_idx != 2 else 2)

    with pytest.raises(ValueError):
        with Control(eng, control):
            AddQuantum | (qureg_a, qureg_b)


@pytest.mark.parametrize('inverse', (False, True))
@pytest.mark.parametrize('qubit_idx', (0, 1, 2, 3))
def test_quantum_conditional_add_carry_size_mismatch(eng, qubit_idx, inverse):
    qureg_a = eng.allocate_qureg(4 if qubit_idx != 0 else 3)
    qureg_b = eng.allocate_qureg(4 if qubit_idx != 1 else 3)
    qureg_c = eng.allocate_qureg(2 if qubit_idx != 2 else 3)
    control = eng.allocate_qureg(1 if qubit_idx != 3 else 2)

    with pytest.raises(ValueError):
        with Control(eng, control):
            if inverse:
                with Dagger(eng):
                    AddQuantum | (qureg_a, qureg_b, qureg_c)
            else:
                AddQuantum | (qureg_a, qureg_b, qureg_c)


def test_quantum_adder(eng):
    qureg_a = eng.allocate_qureg(4)
    qureg_b = eng.allocate_qureg(4)
    control_qubit = eng.allocate_qubit()

    init(eng, qureg_a, 2)
    init(eng, qureg_b, 1)
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 1, 0, 0], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 0, 0, 0], qureg_b))

    with Control(eng, control_qubit):
        AddQuantum | (qureg_a, qureg_b)

    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 1, 0, 0], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 0, 0, 0], qureg_b))

    X | control_qubit

    with Control(eng, control_qubit):
        AddQuantum | (qureg_a, qureg_b)

    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 1, 0, 0], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 0, 0], qureg_b))

    init(eng, qureg_a, 2)  # reset
    init(eng, qureg_b, 3)  # reset

    c = eng.allocate_qubit()
    init(eng, qureg_a, 15)
    init(eng, qureg_b, 15)

    AddQuantum | (qureg_a, qureg_b, c)

    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 1, 1], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 1, 1, 1], qureg_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1], c))

    with Compute(eng):
        with Control(eng, control_qubit):
            AddQuantum | (qureg_a, qureg_b)
    Uncompute(eng)

    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 1, 1], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 1, 1, 1], qureg_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1], c))

    AddQuantum | (qureg_a, qureg_b)

    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 1, 1], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 0, 1, 1], qureg_b))

    with Compute(eng):
        AddQuantum | (qureg_a, qureg_b)
    Uncompute(eng)

    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 1, 1], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 0, 1, 1], qureg_b))

    d = eng.allocate_qureg(2)

    with Compute(eng):
        with Control(eng, control_qubit):
            AddQuantum | (qureg_a, qureg_b, d)
    Uncompute(eng)

    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 1, 1], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 0, 1, 1], qureg_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0], d))

    All(Measure) | qureg_b
    Measure | c


def test_quantumsubtraction(eng):
    qureg_a = eng.allocate_qureg(4)
    qureg_b = eng.allocate_qureg(4)
    control_qubit = eng.allocate_qubit()

    init(eng, qureg_a, 5)
    init(eng, qureg_b, 7)

    X | control_qubit
    with Control(eng, control_qubit):
        SubtractQuantum | (qureg_a, qureg_b)

    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 0, 1, 0, 0], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 1, 0, 0, 0], qureg_b))

    # Size mismatch
    with pytest.raises(ValueError):
        SubtractQuantum | (qureg_a, qureg_b[:-1])
        eng.flush()

    init(eng, qureg_a, 5)  # reset
    init(eng, qureg_b, 2)  # reset

    init(eng, qureg_a, 5)
    init(eng, qureg_b, 3)

    SubtractQuantum | (qureg_a, qureg_b)

    print_all_probabilities(eng, qureg_b)
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 0, 1, 0, 0], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 1, 1, 1, 0], qureg_b))

    init(eng, qureg_a, 5)  # reset
    init(eng, qureg_b, 14)  # reset
    init(eng, qureg_a, 5)
    init(eng, qureg_b, 3)

    with Compute(eng):
        SubtractQuantum | (qureg_a, qureg_b)
    Uncompute(eng)

    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 0, 1, 0, 0], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 0, 0, 0], qureg_b))
    All(Measure) | qureg_a
    All(Measure) | qureg_b


def test_comparator(eng):
    qureg_a = eng.allocate_qureg(3)
    qureg_b = eng.allocate_qureg(3)
    compare_qubit = eng.allocate_qubit()

    init(eng, qureg_a, 5)
    init(eng, qureg_b, 3)

    ComparatorQuantum | (qureg_a, qureg_b, compare_qubit)

    assert 1.0 == pytest.approx(eng.backend.get_probability([1], compare_qubit))

    # Size mismatch in qubit registers
    with pytest.raises(ValueError):
        ComparatorQuantum | (qureg_a, qureg_b[:-1], compare_qubit)

    # Only single qubit for compare qubit
    with pytest.raises(ValueError):
        ComparatorQuantum | (qureg_a, qureg_b, [*compare_qubit, *compare_qubit])

    All(Measure) | qureg_a
    All(Measure) | qureg_b
    Measure | compare_qubit


@pytest.mark.parametrize('qubit_idx', (0, 1, 2))
def test_quantumdivision_size_mismatch(eng, qubit_idx):
    qureg_a = eng.allocate_qureg(4 if qubit_idx != 0 else 3)
    qureg_b = eng.allocate_qureg(4 if qubit_idx != 1 else 3)
    qureg_c = eng.allocate_qureg(4 if qubit_idx != 2 else 3)

    with pytest.raises(ValueError):
        DivideQuantum | (qureg_a, qureg_b, qureg_c)

    All(Measure) | qureg_a
    All(Measure) | qureg_b
    All(Measure) | qureg_c


@pytest.mark.parametrize('qubit_idx', (0, 1, 2))
def test_quantumdivision_size_mismatch_inverse(eng, qubit_idx):
    qureg_a = eng.allocate_qureg(4 if qubit_idx != 0 else 3)
    qureg_b = eng.allocate_qureg(4 if qubit_idx != 1 else 3)
    qureg_c = eng.allocate_qureg(4 if qubit_idx != 2 else 3)

    with pytest.raises(ValueError):
        with Dagger(eng):
            DivideQuantum | (qureg_a, qureg_b, qureg_c)

    All(Measure) | qureg_a
    All(Measure) | qureg_b
    All(Measure) | qureg_c


def test_quantumdivision(eng):
    qureg_a = eng.allocate_qureg(4)
    qureg_b = eng.allocate_qureg(4)
    qureg_c = eng.allocate_qureg(4)

    init(eng, qureg_a, 10)
    init(eng, qureg_c, 3)

    DivideQuantum | (qureg_a, qureg_b, qureg_c)

    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 0, 0, 0], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 0, 0], qureg_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 0, 0], qureg_c))

    All(Measure) | qureg_a
    All(Measure) | qureg_b
    All(Measure) | qureg_c

    init(eng, qureg_a, 1)  # reset
    init(eng, qureg_b, 3)  # reset

    init(eng, qureg_a, 11)

    with Compute(eng):
        DivideQuantum | (qureg_a, qureg_b, qureg_c)
    Uncompute(eng)

    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 0, 1], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 0, 0], qureg_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 0, 0], qureg_c))

    All(Measure) | qureg_a
    All(Measure) | qureg_b
    All(Measure) | qureg_c


@pytest.mark.parametrize('qubit_idx', (0, 1, 2))
def test_quantummultiplication_size_mismatch(eng, qubit_idx):
    qureg_a = eng.allocate_qureg(3 if qubit_idx != 0 else 2)
    qureg_b = eng.allocate_qureg(3 if qubit_idx != 1 else 2)
    qureg_c = eng.allocate_qureg(7 if qubit_idx != 2 else 6)

    with pytest.raises(ValueError):
        MultiplyQuantum | (qureg_a, qureg_b, qureg_c)

    All(Measure) | qureg_a
    All(Measure) | qureg_b
    All(Measure) | qureg_c


@pytest.mark.parametrize('qubit_idx', (0, 1, 2))
def test_quantummultiplication_size_mismatch_inverse(eng, qubit_idx):
    qureg_a = eng.allocate_qureg(4 if qubit_idx != 0 else 3)
    qureg_b = eng.allocate_qureg(4 if qubit_idx != 1 else 3)
    qureg_c = eng.allocate_qureg(4 if qubit_idx != 2 else 3)

    with pytest.raises(ValueError):
        with Dagger(eng):
            MultiplyQuantum | (qureg_a, qureg_b, qureg_c)

    All(Measure) | qureg_a
    All(Measure) | qureg_b
    All(Measure) | qureg_c


def test_quantummultiplication(eng):
    qureg_a = eng.allocate_qureg(3)
    qureg_b = eng.allocate_qureg(3)
    qureg_c = eng.allocate_qureg(7)

    init(eng, qureg_a, 7)
    init(eng, qureg_b, 3)

    MultiplyQuantum | (qureg_a, qureg_b, qureg_c)

    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 1], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 0], qureg_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 0, 1, 0, 1, 0, 0], qureg_c))

    All(Measure) | qureg_a
    All(Measure) | qureg_b
    All(Measure) | qureg_c

    init(eng, qureg_a, 7)
    init(eng, qureg_b, 3)
    init(eng, qureg_c, 21)

    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 0, 0, 0, 0, 0], qureg_c))
    init(eng, qureg_a, 2)
    init(eng, qureg_b, 3)

    with Compute(eng):
        MultiplyQuantum | (qureg_a, qureg_b, qureg_c)
    Uncompute(eng)

    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 1, 0], qureg_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 0], qureg_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 0, 0, 0, 0, 0], qureg_c))



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ==============================================================================
# This file contains:
#
# - Helper fixtures:
#   * arntask
#   * creds
#   * s3_folder
#   * info
#   * results_json
#   * results_dict
#   * res_completed
#   * search_value
#   * device_value
#   * devicelist_result
# - Setup fixtures for specific tests:
#   * show_devices_setup
#   * retrieve_setup
#   * retrieve_devicetypes_setup
#   * send_too_many_setup
#   * real_device_online_setup
# ==============================================================================

from io import StringIO
import json
import pytest

try:
    from botocore.response import StreamingBody
except ImportError:

    class StreamingBody:
        def __init__(self, raw_stream, content_length):
            pass


# ==============================================================================


@pytest.fixture
def arntask():
    return 'arn:aws:braket:us-east-1:id:taskuuid'


@pytest.fixture
def creds():
    return {
        'AWS_ACCESS_KEY_ID': 'aws_access_key_id',
        'AWS_SECRET_KEY': 'aws_secret_key',
    }


@pytest.fixture
def s3_folder():
    return ['S3Bucket', "S3Directory"]


@pytest.fixture
def info():
    return {
        'circuit': '{"braketSchemaHeader":'
        '{"name": "braket.ir.jaqcd.program", "version": "1"}, '
        '"results": [], "basis_rotation_instructions": [], '
        '"instructions": [{"target": 0, "type": "h"}, {\
            "target": 1, "type": "h"}, {\
                "control": 1, "target": 2, "type": "cnot"}]}',
        'nq': 10,
        'shots': 1,
        'backend': {'name': 'name2'},
    }


@pytest.fixture
def results_json():
    return json.dumps(
        {
            "braketSchemaHeader": {
                "name": "braket.task_result.gate_model_task_result",
                "version": "1",
            },
            "measurementProbabilities": {
                "000": 0.1,
                "010": 0.4,
                "110": 0.1,
                "001": 0.1,
                "111": 0.3,
            },
            "measuredQubits": [0, 1, 2],
        }
    )


@pytest.fixture
def results_dict(results_json):
    body = StreamingBody(StringIO(results_json), len(results_json))
    return {
        'ResponseMetadata': {
            'RequestId': 'CF4CAA48CC18836C',
            'HTTPHeaders': {},
        },
        'Body': body,
    }


@pytest.fixture
def res_completed():
    return {"000": 0.1, "010": 0.4, "110": 0.1, "001": 0.1, "111": 0.3}


@pytest.fixture
def search_value():
    return {
        "devices": [
            {
                "deviceArn": "arn1",
                "deviceName": "name1",
                "deviceType": "SIMULATOR",
                "deviceStatus": "ONLINE",
                "providerName": "pname1",
            },
            {
                "deviceArn": "arn2",
                "deviceName": "name2",
                "deviceType": "QPU",
                "deviceStatus": "OFFLINE",
                "providerName": "pname1",
            },
            {
                "deviceArn": "arn3",
                "deviceName": "name3",
                "deviceType": "QPU",
                "deviceStatus": "ONLINE",
                "providerName": "pname2",
            },
            {
                "deviceArn": "invalid",
                "deviceName": "invalid",
                "deviceType": "BLABLA",
                "deviceStatus": "ONLINE",
                "providerName": "pname3",
            },
        ]
    }


@pytest.fixture
def device_value_devicecapabilities():
    return json.dumps(
        {
            "braketSchemaHeader": {
                "name": "braket.device_schema.rigetti.rigetti_device_capabilities",
                "version": "1",
            },
            "service": {
                "executionWindows": [
                    {
                        "executionDay": "Everyday",
                        "windowStartHour": "11:00",
                        "windowEndHour": "12:00",
                    }
                ],
                "shotsRange": [1, 10],
                "deviceLocation": "us-east-1",
            },
            "action": {
                "braket.ir.jaqcd.program": {
                    "actionType": "braket.ir.jaqcd.program",
                    "version": ["1"],
                    "supportedOperations": ["H"],
                }
            },
            "paradigm": {
                "qubitCount": 30,
                "nativeGateSet": ["ccnot", "cy"],
                "connectivity": {
                    "fullyConnected": False,
                    "connectivityGraph": {"1": ["2", "3"]},
                },
            },
            "deviceParameters": {
                "properties": {
                    "braketSchemaHeader": {
                        "const": {
                            "name": "braket.device_schema.rigetti.rigetti_device_parameters",
                            "version": "1",
                        }
                    }
                },
                "definitions": {
                    "GateModelParameters": {
                        "properties": {
                            "braketSchemaHeader": {
                                "const": {
                                    "name": "braket.device_schema.gate_model_parameters",
                                    "version": "1",
                                }
                            }
                        }
                    }
                },
            },
        }
    )


@pytest.fixture
def device_value(device_value_devicecapabilities):
    return {
        "deviceName": "Aspen-8",
        "deviceType": "QPU",
        "providerName": "provider1",
        "deviceStatus": "OFFLINE",
        "deviceCapabilities": device_value_devicecapabilities,
    }


@pytest.fixture
def devicelist_result():
    return {
        'name1': {
            'coupling_map': {},
            'deviceArn': 'arn1',
            'location': 'us-east-1',
            'nq': 30,
            'version': '1',
            'deviceParameters': {
                'name': 'braket.device_schema.rigetti.rigetti_device_parameters',
                'version': '1',
            },
            'deviceModelParameters': {
                'name': 'braket.device_schema.gate_model_parameters',
                'version': '1',
            },
        },
        'name2': {
            'coupling_map': {'1': ['2', '3']},
            'deviceArn': 'arn2',
            'location': 'us-east-1',
            'nq': 30,
            'version': '1',
            'deviceParameters': {
                'name': 'braket.device_schema.rigetti.rigetti_device_parameters',
                'version': '1',
            },
            'deviceModelParameters': {
                'name': 'braket.device_schema.gate_model_parameters',
                'version': '1',
            },
        },
        'name3': {
            'coupling_map': {'1': ['2', '3']},
            'deviceArn': 'arn3',
            'location': 'us-east-1',
            'nq': 30,
            'version': '1',
            'deviceParameters': {
                'name': 'braket.device_schema.rigetti.rigetti_device_parameters',
                'version': '1',
            },
            'deviceModelParameters': {
                'name': 'braket.device_schema.gate_model_parameters',
                'version': '1',
            },
        },
    }


# ==============================================================================


@pytest.fixture
def show_devices_setup(creds, search_value, device_value, devicelist_result):
    return creds, search_value, device_value, devicelist_result


@pytest.fixture
def retrieve_setup(arntask, creds, device_value, res_completed, results_dict):
    return arntask, creds, device_value, res_completed, results_dict


@pytest.fixture(params=["qpu", "sim"])
def retrieve_devicetypes_setup(request, arntask, creds, results_json, device_value_devicecapabilities):
    if request.param == "qpu":
        body_qpu = StreamingBody(StringIO(results_json), len(results_json))
        results_dict = {
            'ResponseMetadata': {
                'RequestId': 'CF4CAA48CC18836C',
                'HTTPHeaders': {},
            },
            'Body': body_qpu,
        }

        device_value = {
            "deviceName": "Aspen-8",
            "deviceType": "QPU",
            "providerName": "provider1",
            "deviceStatus": "OFFLINE",
            "deviceCapabilities": device_value_devicecapabilities,
        }

        res_completed = {"000": 0.1, "010": 0.4, "110": 0.1, "001": 0.1, "111": 0.3}
    else:
        results_json_simulator = json.dumps(
            {
                "braketSchemaHeader": {
                    "name": "braket.task_result.gate_model_task_result",
                    "version": "1",
                },
                "measurements": [
                    [0, 0],
                    [0, 1],
                    [1, 1],
                    [0, 1],
                    [0, 1],
                    [1, 1],
                    [1, 1],
                    [1, 1],
                    [1, 1],
                    [1, 1],
                ],
                "measuredQubits": [0, 1],
            }
        )
        body_simulator = StreamingBody(StringIO(results_json_simulator), len(results_json_simulator))
        results_dict = {
            'ResponseMetadata': {
                'RequestId': 'CF4CAA48CC18836C',
                'HTTPHeaders': {},
            },
            'Body': body_simulator,
        }

        device_value = {
            "deviceName": "SV1",
            "deviceType": "SIMULATOR",
            "providerName": "providerA",
            "deviceStatus": "ONLINE",
            "deviceCapabilities": device_value_devicecapabilities,
        }

        res_completed = {"00": 0.1, "01": 0.3, "11": 0.6}
    return arntask, creds, device_value, results_dict, res_completed


@pytest.fixture
def send_too_many_setup(creds, s3_folder, search_value, device_value):
    info_too_much = {
        'circuit': '{"braketSchemaHeader":'
        '{"name": "braket.ir.jaqcd.program", "version": "1"}, '
        '"results": [], "basis_rotation_instructions": [], '
        '"instructions": [{"target": 0, "type": "h"}, {\
            "target": 1, "type": "h"}, {\
                "control": 1, "target": 2, "type": "cnot"}]}',
        'nq': 100,
        'shots': 1,
        'backend': {'name': 'name2'},
    }
    return creds, s3_folder, search_value, device_value, info_too_much


@pytest.fixture
def real_device_online_setup(
    arntask,
    creds,
    s3_folder,
    info,
    search_value,
    device_value,
    res_completed,
    results_json,
):
    qtarntask = {'quantumTaskArn': arntask}
    body = StreamingBody(StringIO(results_json), len(results_json))
    results_dict = {
        'ResponseMetadata': {
            'RequestId': 'CF4CAA48CC18836C',
            'HTTPHeaders': {},
        },
        'Body': body,
    }

    return (
        qtarntask,
        creds,
        s3_folder,
        info,
        search_value,
        device_value,
        res_completed,
        results_dict,
    )


@pytest.fixture
def send_that_error_setup(creds, s3_folder, info, search_value, device_value):
    return creds, s3_folder, info, search_value, device_value



# -*- coding: utf-8 -*-
#   Copyright 2020 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.libs.math._gates.py."""

import pytest

from projectq.cengines import (
    MainEngine,
    TagRemover,
    AutoReplacer,
    InstructionFilter,
    DecompositionRuleSet,
)
from projectq.meta import Control, Compute, Uncompute
from projectq.ops import All, Measure, X, BasicMathGate, ClassicalInstructionGate
import projectq.setups.decompositions

import projectq.libs.math
from . import (
    AddConstant,
    AddQuantum,
    SubtractQuantum,
    ComparatorQuantum,
    DivideQuantum,
    MultiplyQuantum,
)

from projectq.backends import CommandPrinter


def print_all_probabilities(eng, qureg):
    i = 0
    y = len(qureg)
    while i < (2 ** y):
        qubit_list = [int(x) for x in list(('{0:0b}'.format(i)).zfill(y))]
        qubit_list = qubit_list[::-1]
        prob = eng.backend.get_probability(qubit_list, qureg)
        if prob != 0.0:
            print(prob, qubit_list, i)

        i += 1


def _eng_emulation():
    # Only decomposing native ProjectQ gates
    # -> using emulation for gates in projectq.libs.math
    rule_set = DecompositionRuleSet(modules=[projectq.setups.decompositions])
    eng = MainEngine(
        engine_list=[
            TagRemover(),
            AutoReplacer(rule_set),
            TagRemover(),
            CommandPrinter(),
        ],
        verbose=True,
    )
    return eng


def _eng_decomp():
    def no_math_emulation(eng, cmd):
        if isinstance(cmd.gate, BasicMathGate):
            return False
        if isinstance(cmd.gate, ClassicalInstructionGate):
            return True
        try:
            return len(cmd.gate.matrix) > 0
        except AttributeError:
            return False

    rule_set = DecompositionRuleSet(modules=[projectq.libs.math, projectq.setups.decompositions.qft2crandhadamard])
    eng = MainEngine(
        engine_list=[
            TagRemover(),
            AutoReplacer(rule_set),
            InstructionFilter(no_math_emulation),
            TagRemover(),
            CommandPrinter(),
        ]
    )
    return eng


@pytest.fixture(params=['no_decomp', 'full_decomp'])
def eng(request):
    if request.param == 'no_decomp':
        return _eng_emulation()
    elif request.param == 'full_decomp':
        return _eng_decomp()


def test_constant_addition(eng):
    qunum_a = eng.allocate_qureg(5)
    X | qunum_a[2]
    with Compute(eng):
        AddConstant(5) | (qunum_a)

    Uncompute(eng)

    eng.flush()

    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 1, 0, 0], qunum_a))


def test_addition(eng):
    qunum_a = eng.allocate_qureg(5)  # 5-qubit number
    qunum_b = eng.allocate_qureg(5)  # 5-qubit number
    carry_bit = eng.allocate_qubit()
    X | qunum_a[2]  # qunum_a is now equal to 4
    X | qunum_b[3]  # qunum_b is now equal to 8
    AddQuantum | (qunum_a, qunum_b, carry_bit)

    eng.flush()

    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 1, 0, 0], qunum_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 1, 1, 0], qunum_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0], carry_bit))


def test_inverse_addition(eng):
    qunum_a = eng.allocate_qureg(5)
    qunum_b = eng.allocate_qureg(5)
    X | qunum_a[2]
    X | qunum_b[3]
    with Compute(eng):
        AddQuantum | (qunum_a, qunum_b)
    Uncompute(eng)

    eng.flush()

    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 1, 0, 0], qunum_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 0, 1, 0], qunum_b))


def test_inverse_addition_with_control(eng):
    qunum_a = eng.allocate_qureg(5)
    qunum_b = eng.allocate_qureg(5)
    qunum_c = eng.allocate_qubit()
    All(X) | qunum_a
    All(X) | qunum_b
    X | qunum_c
    with Compute(eng):
        with Control(eng, qunum_c):
            AddQuantum | (qunum_a, qunum_b)

    Uncompute(eng)

    eng.flush()

    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 1, 1, 1], qunum_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 1, 1, 1], qunum_b))


def test_addition_with_control(eng):
    qunum_a = eng.allocate_qureg(5)
    qunum_b = eng.allocate_qureg(5)
    control_bit = eng.allocate_qubit()
    X | qunum_a[1]  # qunum_a is now equal to 2
    X | qunum_b[4]  # qunum_b is now equal to 16
    X | control_bit
    with Control(eng, control_bit):
        AddQuantum | (qunum_a, qunum_b)

    eng.flush()

    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 1, 0, 0, 0], qunum_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 1, 0, 0, 1], qunum_b))


def test_addition_with_control_carry(eng):
    qunum_a = eng.allocate_qureg(4)  # 4-qubit number
    qunum_b = eng.allocate_qureg(4)  # 4-qubit number
    control_bit = eng.allocate_qubit()
    qunum_c = eng.allocate_qureg(2)

    X | qunum_a[1]  # qunum is now equal to 2
    All(X) | qunum_b[0:4]  # qunum is now equal to 15
    X | control_bit

    with Control(eng, control_bit):
        AddQuantum | (qunum_a, qunum_b, qunum_c)
    # qunum_a and ctrl don't change, qunum_b and qunum_c are now both equal
    # to 1 so in binary together 10001 (2 + 15 = 17)

    eng.flush()

    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 1, 0, 0], qunum_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 0, 0, 0], qunum_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1], control_bit))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 0], qunum_c))

    All(Measure) | qunum_a
    All(Measure) | qunum_b


def test_inverse_addition_with_control_carry(eng):
    qunum_a = eng.allocate_qureg(4)
    qunum_b = eng.allocate_qureg(4)

    control_bit = eng.allocate_qubit()
    qunum_c = eng.allocate_qureg(2)

    X | qunum_a[1]
    All(X) | qunum_b[0:4]
    X | control_bit
    with Compute(eng):
        with Control(eng, control_bit):
            AddQuantum | (qunum_a, qunum_b, qunum_c)
    Uncompute(eng)

    eng.flush()

    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 1, 0, 0], qunum_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 1, 1, 1], qunum_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1], control_bit))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0], qunum_c))

    All(Measure) | qunum_a
    All(Measure) | qunum_b
    Measure | control_bit
    All(Measure) | qunum_c


def test_subtraction(eng):
    qunum_a = eng.allocate_qureg(5)
    qunum_b = eng.allocate_qureg(5)

    X | qunum_a[2]
    X | qunum_b[3]

    SubtractQuantum | (qunum_a, qunum_b)

    eng.flush()

    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 1, 0, 0], qunum_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 1, 0, 0], qunum_b))


def test_inverse_subtraction(eng):
    qunum_a = eng.allocate_qureg(5)
    qunum_b = eng.allocate_qureg(5)

    X | qunum_a[2]
    X | qunum_b[3]

    with Compute(eng):
        SubtractQuantum | (qunum_a, qunum_b)
    Uncompute(eng)

    eng.flush()

    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 1, 0, 0], qunum_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 0, 1, 0], qunum_b))


def test_comparator(eng):
    qunum_a = eng.allocate_qureg(5)  # 5-qubit number
    qunum_b = eng.allocate_qureg(5)  # 5-qubit number
    compare_bit = eng.allocate_qubit()
    X | qunum_a[4]  # qunum_a is now equal to 16
    X | qunum_b[3]  # qunum_b is now equal to 8

    ComparatorQuantum | (qunum_a, qunum_b, compare_bit)

    eng.flush()
    print_all_probabilities(eng, qunum_a)
    print_all_probabilities(eng, qunum_b)
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 0, 0, 1], qunum_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 0, 1, 0], qunum_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([1], compare_bit))


def test_division(eng):
    qunum_a = eng.allocate_qureg(5)
    qunum_b = eng.allocate_qureg(5)
    qunum_c = eng.allocate_qureg(5)

    All(X) | [qunum_a[0], qunum_a[3]]  # qunum_a is now equal to 9
    X | qunum_c[2]  # qunum_c is now 4

    DivideQuantum | (qunum_a, qunum_b, qunum_c)
    eng.flush()

    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 0, 0, 0, 0], qunum_a))  # remainder
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 1, 0, 0, 0], qunum_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 1, 0, 0], qunum_c))


def test_inverse_division(eng):
    qunum_a = eng.allocate_qureg(5)
    qunum_b = eng.allocate_qureg(5)
    qunum_c = eng.allocate_qureg(5)

    All(X) | [qunum_a[0], qunum_a[3]]
    X | qunum_c[2]

    with Compute(eng):
        DivideQuantum | (qunum_a, qunum_b, qunum_c)
    Uncompute(eng)
    eng.flush()

    assert 1.0 == pytest.approx(eng.backend.get_probability([1, 0, 0, 1, 0], qunum_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 0, 0, 0], qunum_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 1, 0, 0], qunum_c))


def test_multiplication(eng):
    qunum_a = eng.allocate_qureg(4)
    qunum_b = eng.allocate_qureg(4)
    qunum_c = eng.allocate_qureg(9)
    X | qunum_a[2]  # qunum_a is now 4
    X | qunum_b[3]  # qunum_b is now 8
    MultiplyQuantum | (qunum_a, qunum_b, qunum_c)
    # qunum_a remains 4 and qunum_b remains 8, qunum_c is now equal to 32

    eng.flush()

    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 1, 0, 0], qunum_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 0, 1, 0], qunum_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 0, 0, 0, 1, 0, 0, 0], qunum_c))


def test_inverse_multiplication(eng):
    qunum_a = eng.allocate_qureg(4)
    qunum_b = eng.allocate_qureg(4)
    qunum_c = eng.allocate_qureg(9)
    X | qunum_a[2]  # qunum_a is now 4
    X | qunum_b[3]  # qunum_b is now 8
    with Compute(eng):
        MultiplyQuantum | (qunum_a, qunum_b, qunum_c)
    Uncompute(eng)

    eng.flush()

    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 1, 0, 0], qunum_a))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 0, 1, 0], qunum_b))
    assert 1.0 == pytest.approx(eng.backend.get_probability([0, 0, 0, 0, 0, 0, 0, 0, 0], qunum_c))



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.cengines._manualmapper.py."""

from projectq import MainEngine
from projectq.cengines import DummyEngine
from projectq.ops import H, Measure, All
from projectq.meta import LogicalQubitIDTag

from projectq.cengines import ManualMapper


def test_manualmapper_mapping():
    backend = DummyEngine(save_commands=True)

    def mapping(qubit_id):
        return (qubit_id + 1) & 1

    eng = MainEngine(backend=backend, engine_list=[ManualMapper(mapping)])
    qb0 = eng.allocate_qubit()
    qb1 = eng.allocate_qubit()
    H | qb0
    H | qb1
    All(Measure) | (qb0 + qb1)
    eng.flush()

    num_measurements = 0
    for cmd in backend.received_commands:
        if cmd.gate == Measure:
            tag = LogicalQubitIDTag(mapping(cmd.qubits[0][0].id))
            assert tag in cmd.tags
            wrong_tag = LogicalQubitIDTag(cmd.qubits[0][0].id)
            assert wrong_tag not in cmd.tags
            num_measurements += 1
    assert num_measurements == 2



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Register a decomposition to replace turn negatively controlled qubits into positively controlled qubits by applying X
gates.
"""

from copy import deepcopy
from projectq.cengines import DecompositionRule
from projectq.meta import Compute, Uncompute, has_negative_control
from projectq.ops import BasicGate, X


def _decompose_controlstate(cmd):
    """
    Decompose commands with control qubits in negative state (ie. control
    qubits with state '0' instead of '1')
    """
    with Compute(cmd.engine):
        for state, ctrl in zip(cmd.control_state, cmd.control_qubits):
            if state == '0':
                X | ctrl

    # Resend the command with the `control_state` cleared
    cmd.ctrl_state = '1' * len(cmd.control_state)
    orig_engine = cmd.engine
    cmd.engine.receive([deepcopy(cmd)])  # NB: deepcopy required here to workaround infinite recursion detection
    Uncompute(orig_engine)


#: Decomposition rules
all_defined_decomposition_rules = [DecompositionRule(BasicGate, _decompose_controlstate, has_negative_control)]



# -*- coding: utf-8 -*-
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.meta._control.py"""
import pytest

from projectq import MainEngine
from projectq.cengines import DummyEngine
from projectq.ops import Command, H, Rx, CtrlAll, X, IncompatibleControlState
from projectq.meta import DirtyQubitTag, ComputeTag, UncomputeTag, Compute, Uncompute

from projectq.meta import _control
from projectq.types import WeakQubitRef


def test_canonical_representation():
    assert _control.canonical_ctrl_state(0, 0) == ''
    for num_qubits in range(4):
        assert _control.canonical_ctrl_state(0, num_qubits) == '0' * num_qubits

    num_qubits = 4
    for i in range(2 ** num_qubits):
        state = '{0:0b}'.format(i).zfill(num_qubits)
        assert _control.canonical_ctrl_state(i, num_qubits) == state[::-1]
        assert _control.canonical_ctrl_state(state, num_qubits) == state

    for num_qubits in range(10):
        assert _control.canonical_ctrl_state(CtrlAll.Zero, num_qubits) == '0' * num_qubits
        assert _control.canonical_ctrl_state(CtrlAll.One, num_qubits) == '1' * num_qubits

    with pytest.raises(TypeError):
        _control.canonical_ctrl_state(1.1, 2)

    with pytest.raises(ValueError):
        _control.canonical_ctrl_state('1', 2)

    with pytest.raises(ValueError):
        _control.canonical_ctrl_state('11111', 2)

    with pytest.raises(ValueError):
        _control.canonical_ctrl_state('1a', 2)

    with pytest.raises(ValueError):
        _control.canonical_ctrl_state(4, 2)


def test_has_negative_control():
    qubit0 = WeakQubitRef(None, 0)
    qubit1 = WeakQubitRef(None, 0)
    qubit2 = WeakQubitRef(None, 0)
    qubit3 = WeakQubitRef(None, 0)
    assert not _control.has_negative_control(Command(None, H, ([qubit0],)))
    assert not _control.has_negative_control(Command(None, H, ([qubit0],), [qubit1]))
    assert not _control.has_negative_control(Command(None, H, ([qubit0],), [qubit1], control_state=CtrlAll.One))
    assert _control.has_negative_control(Command(None, H, ([qubit0],), [qubit1], control_state=CtrlAll.Zero))
    assert _control.has_negative_control(
        Command(None, H, ([qubit0],), [qubit1, qubit2, qubit3], control_state=CtrlAll.Zero)
    )
    assert not _control.has_negative_control(
        Command(None, H, ([qubit0],), [qubit1, qubit2, qubit3], control_state='111')
    )
    assert _control.has_negative_control(Command(None, H, ([qubit0],), [qubit1, qubit2, qubit3], control_state='101'))


def test_control_engine_has_compute_tag():
    eng = MainEngine(backend=DummyEngine(), engine_list=[DummyEngine()])
    qubit = eng.allocate_qubit()
    test_cmd0 = Command(eng, H, (qubit,))
    test_cmd1 = Command(eng, H, (qubit,))
    test_cmd2 = Command(eng, H, (qubit,))
    test_cmd0.tags = [DirtyQubitTag(), ComputeTag(), DirtyQubitTag()]
    test_cmd1.tags = [DirtyQubitTag(), UncomputeTag(), DirtyQubitTag()]
    test_cmd2.tags = [DirtyQubitTag()]
    assert _control._has_compute_uncompute_tag(test_cmd0)
    assert _control._has_compute_uncompute_tag(test_cmd1)
    assert not _control._has_compute_uncompute_tag(test_cmd2)


def test_control():
    backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=backend, engine_list=[DummyEngine()])
    qureg = eng.allocate_qureg(2)
    with _control.Control(eng, qureg):
        qubit = eng.allocate_qubit()
        with Compute(eng):
            Rx(0.5) | qubit
        H | qubit
        Uncompute(eng)
    with _control.Control(eng, qureg[0]):
        H | qubit
    eng.flush()
    assert len(backend.received_commands) == 8
    assert len(backend.received_commands[0].control_qubits) == 0
    assert len(backend.received_commands[1].control_qubits) == 0
    assert len(backend.received_commands[2].control_qubits) == 0
    assert len(backend.received_commands[3].control_qubits) == 0
    assert len(backend.received_commands[4].control_qubits) == 2
    assert len(backend.received_commands[5].control_qubits) == 0
    assert len(backend.received_commands[6].control_qubits) == 1
    assert len(backend.received_commands[7].control_qubits) == 0
    assert backend.received_commands[4].control_qubits[0].id == qureg[0].id
    assert backend.received_commands[4].control_qubits[1].id == qureg[1].id
    assert backend.received_commands[6].control_qubits[0].id == qureg[0].id

    with pytest.raises(TypeError):
        _control.Control(eng, (qureg[0], qureg[1]))


def test_control_state():
    backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=backend, engine_list=[DummyEngine()])

    qureg = eng.allocate_qureg(3)
    xreg = eng.allocate_qureg(3)
    X | qureg[1]
    with _control.Control(eng, qureg[0], '0'):
        with Compute(eng):
            X | xreg[0]

        X | xreg[1]
        Uncompute(eng)

    with _control.Control(eng, qureg[1:], 2):
        X | xreg[2]
    eng.flush()

    assert len(backend.received_commands) == 6 + 5 + 1
    assert len(backend.received_commands[0].control_qubits) == 0
    assert len(backend.received_commands[1].control_qubits) == 0
    assert len(backend.received_commands[2].control_qubits) == 0
    assert len(backend.received_commands[3].control_qubits) == 0
    assert len(backend.received_commands[4].control_qubits) == 0
    assert len(backend.received_commands[5].control_qubits) == 0

    assert len(backend.received_commands[6].control_qubits) == 0
    assert len(backend.received_commands[7].control_qubits) == 0
    assert len(backend.received_commands[8].control_qubits) == 1
    assert len(backend.received_commands[9].control_qubits) == 0
    assert len(backend.received_commands[10].control_qubits) == 2

    assert len(backend.received_commands[11].control_qubits) == 0

    assert backend.received_commands[8].control_qubits[0].id == qureg[0].id
    assert backend.received_commands[8].control_state == '0'
    assert backend.received_commands[10].control_qubits[0].id == qureg[1].id
    assert backend.received_commands[10].control_qubits[1].id == qureg[2].id
    assert backend.received_commands[10].control_state == '01'

    assert _control.has_negative_control(backend.received_commands[8])
    assert _control.has_negative_control(backend.received_commands[10])


def test_control_state_contradiction():
    backend = DummyEngine(save_commands=True)
    eng = MainEngine(backend=backend, engine_list=[DummyEngine()])
    qureg = eng.allocate_qureg(1)
    with pytest.raises(IncompatibleControlState):
        with _control.Control(eng, qureg[0], '0'):
            qubit = eng.allocate_qubit()
            with _control.Control(eng, qureg[0], '1'):
                H | qubit
    eng.flush()



# -*- coding: utf-8 -*-
#   Copyright 2020 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.setup.aqt."""

import pytest


def test_aqt_mapper_in_cengines(monkeypatch):
    import projectq.setups.aqt

    def mock_show_devices(*args, **kwargs):
        connections = set(
            [
                (0, 1),
                (1, 0),
                (1, 2),
                (1, 3),
                (1, 4),
                (2, 1),
                (2, 3),
                (2, 4),
                (3, 1),
                (3, 4),
                (4, 3),
            ]
        )
        return {'aqt_simulator': {'coupling_map': connections, 'version': '0.0.0', 'nq': 32}}

    monkeypatch.setattr(projectq.setups.aqt, "show_devices", mock_show_devices)
    engines_simulator = projectq.setups.aqt.get_engine_list(device='aqt_simulator')
    assert len(engines_simulator) == 13


def test_aqt_errors(monkeypatch):
    import projectq.setups.aqt

    def mock_show_devices(*args, **kwargs):
        connections = set(
            [
                (0, 1),
                (1, 0),
                (1, 2),
                (1, 3),
                (1, 4),
                (2, 1),
                (2, 3),
                (2, 4),
                (3, 1),
                (3, 4),
                (4, 3),
            ]
        )
        return {'aqt_imaginary': {'coupling_map': connections, 'version': '0.0.0', 'nq': 6}}

    monkeypatch.setattr(projectq.setups.aqt, "show_devices", mock_show_devices)
    with pytest.raises(projectq.setups.aqt.DeviceOfflineError):
        projectq.setups.aqt.get_engine_list(device='simulator')
    with pytest.raises(projectq.setups.aqt.DeviceNotHandledError):
        projectq.setups.aqt.get_engine_list(device='aqt_imaginary')



# -*- coding: utf-8 -*-
#   Copyright 2021 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the 'License');
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an 'AS IS' BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.backends._ionq._ionq_http_client.py."""

from unittest import mock

import pytest
import requests
from requests.compat import urljoin

from projectq.backends._ionq import _ionq_http_client
from projectq.backends._ionq._ionq_exc import JobSubmissionError, RequestTimeoutError


# Insure that no HTTP request can be made in all tests in this module
@pytest.fixture(autouse=True)
def no_requests(monkeypatch):
    monkeypatch.delattr('requests.sessions.Session.request')


_api_url = 'https://api.ionq.co/v0.1/jobs/'


def test_authenticate():
    ionq_session = _ionq_http_client.IonQ()
    ionq_session.authenticate('NotNone')
    assert 'Authorization' in ionq_session.headers
    assert ionq_session.token == 'NotNone'
    assert ionq_session.headers['Authorization'] == 'apiKey NotNone'


def test_authenticate_prompt_requires_token(monkeypatch):
    def user_password_input(prompt):
        if prompt == 'IonQ apiKey > ':
            return ''

    monkeypatch.setattr('getpass.getpass', user_password_input)
    ionq_session = _ionq_http_client.IonQ()
    with pytest.raises(RuntimeError) as excinfo:
        ionq_session.authenticate()
    assert str(excinfo.value) == 'An authentication token is required!'


def test_is_online():
    ionq_session = _ionq_http_client.IonQ()
    ionq_session.authenticate('not none')
    ionq_session.update_devices_list()
    assert ionq_session.is_online('ionq_simulator')
    assert ionq_session.is_online('ionq_qpu')
    assert not ionq_session.is_online('ionq_unknown')


def test_show_devices():
    device_list = _ionq_http_client.show_devices()
    assert isinstance(device_list, dict)
    for info in device_list.values():
        assert 'nq' in info
        assert 'target' in info


def test_send_too_many_qubits(monkeypatch):
    # Patch the method to give back dummy devices
    def _dummy_update(_self):
        _self.backends = {'dummy': {'nq': 3, 'target': 'dummy'}}

    monkeypatch.setattr(
        _ionq_http_client.IonQ,
        'update_devices_list',
        _dummy_update.__get__(None, _ionq_http_client.IonQ),
    )
    info = {
        'nq': 4,
        'shots': 1,
        'meas_mapped': [2, 3],
        'circuit': [
            {'gate': 'x', 'targets': [0]},
            {'gate': 'x', 'targets': [1]},
            {'controls': [0], 'gate': 'cnot', 'targets': [2]},
            {'controls': [1], 'gate': 'cnot', 'targets': [2]},
            {'controls': [0, 1], 'gate': 'cnot', 'targets': [3]},
        ],
    }
    with pytest.raises(_ionq_http_client.DeviceTooSmall):
        _ionq_http_client.send(
            info,
            device='dummy',
            token='NotNone',
            verbose=True,
        )


def test_send_real_device_online_verbose(monkeypatch):
    # Patch the method to give back dummy devices
    def _dummy_update(_self):
        _self.backends = {'dummy': {'nq': 4, 'target': 'dummy'}}

    monkeypatch.setattr(
        _ionq_http_client.IonQ,
        'update_devices_list',
        _dummy_update.__get__(None, _ionq_http_client.IonQ),
    )
    # What the IonQ JSON API request should look like.
    expected_request = {
        'target': 'dummy',
        'metadata': {'sdk': 'ProjectQ', 'meas_qubit_ids': '[2, 3]'},
        'shots': 1,
        'registers': {'meas_mapped': [2, 3]},
        'lang': 'json',
        'body': {
            'qubits': 4,
            'circuit': [
                {'gate': 'x', 'targets': [0]},
                {'gate': 'x', 'targets': [1]},
                {'controls': [0], 'gate': 'cnot', 'targets': [2]},
                {'controls': [1], 'gate': 'cnot', 'targets': [2]},
                {'controls': [0, 1], 'gate': 'cnot', 'targets': [3]},
            ],
        },
    }

    def mock_post(_self, path, *args, **kwargs):
        assert path == _api_url[:-1]
        assert 'json' in kwargs
        assert expected_request == kwargs['json']
        mock_response = mock.MagicMock()
        mock_response.status_code = 200
        mock_response.json = mock.MagicMock(
            return_value={
                'id': 'new-job-id',
                'status': 'ready',
            }
        )
        return mock_response

    def mock_get(_self, path, *args, **kwargs):
        assert urljoin(_api_url, 'new-job-id') == path
        mock_response = mock.MagicMock()
        mock_response.json = mock.MagicMock(
            return_value={
                'id': 'new-job-id',
                'status': 'completed',
                'qubits': 4,
                'metadata': {'meas_qubit_ids': '[2, 3]'},
                'registers': {'meas_mapped': [2, 3]},
                'data': {
                    'registers': {'meas_mapped': {'2': 1}},
                },
            }
        )
        return mock_response

    monkeypatch.setattr('requests.sessions.Session.post', mock_post)
    monkeypatch.setattr('requests.sessions.Session.get', mock_get)

    def user_password_input(prompt):
        if prompt == 'IonQ apiKey > ':
            return 'NotNone'

    monkeypatch.setattr('getpass.getpass', user_password_input)

    # Code to test:
    info = {
        'nq': 4,
        'shots': 1,
        'meas_mapped': [2, 3],
        'meas_qubit_ids': [2, 3],
        'circuit': [
            {'gate': 'x', 'targets': [0]},
            {'gate': 'x', 'targets': [1]},
            {'controls': [0], 'gate': 'cnot', 'targets': [2]},
            {'controls': [1], 'gate': 'cnot', 'targets': [2]},
            {'controls': [0, 1], 'gate': 'cnot', 'targets': [3]},
        ],
    }
    expected = {
        'nq': 4,
        'output_probs': {'2': 1},
        'meas_mapped': [2, 3],
        'meas_qubit_ids': [2, 3],
    }
    actual = _ionq_http_client.send(info, device='dummy')
    assert expected == actual


@pytest.mark.parametrize(
    'error_type',
    [
        requests.exceptions.HTTPError,
        requests.exceptions.RequestException,
    ],
)
def test_send_requests_errors_are_caught(monkeypatch, error_type):
    # Patch the method to give back dummy devices
    def _dummy_update(_self):
        _self.backends = {'dummy': {'nq': 4, 'target': 'dummy'}}

    monkeypatch.setattr(
        _ionq_http_client.IonQ,
        'update_devices_list',
        _dummy_update.__get__(None, _ionq_http_client.IonQ),
    )
    mock_post = mock.MagicMock(side_effect=error_type())
    monkeypatch.setattr('requests.sessions.Session.post', mock_post)

    def user_password_input(prompt):
        if prompt == 'IonQ apiKey > ':
            return 'NotNone'

    monkeypatch.setattr('getpass.getpass', user_password_input)

    # Code to test:
    info = {
        'nq': 1,
        'shots': 1,
        'meas_mapped': [],
        'meas_qubit_ids': [],
        'circuit': [],
    }
    _ionq_http_client.send(info, device='dummy')
    mock_post.assert_called_once()


def test_send_auth_errors_reraise(monkeypatch):
    # Patch the method to give back dummy devices
    def _dummy_update(_self):
        _self.backends = {'dummy': {'nq': 4, 'target': 'dummy'}}

    monkeypatch.setattr(
        _ionq_http_client.IonQ,
        'update_devices_list',
        _dummy_update.__get__(None, _ionq_http_client.IonQ),
    )

    mock_response = mock.MagicMock()
    mock_response.status_code = 401
    auth_error = requests.exceptions.HTTPError(response=mock_response)
    mock_post = mock.MagicMock(side_effect=auth_error)
    monkeypatch.setattr('requests.sessions.Session.post', mock_post)

    def user_password_input(prompt):
        if prompt == 'IonQ apiKey > ':
            return 'NotNone'

    monkeypatch.setattr('getpass.getpass', user_password_input)

    # Code to test:
    info = {
        'nq': 1,
        'shots': 1,
        'meas_mapped': [],
        'meas_qubit_ids': [],
        'circuit': [],
    }
    with pytest.raises(requests.exceptions.HTTPError) as excinfo:
        _ionq_http_client.send(info, device='dummy')
    mock_post.assert_called_once()
    assert auth_error is excinfo.value


def test_send_bad_requests_reraise(monkeypatch):
    # Patch the method to give back dummy devices
    def _dummy_update(_self):
        _self.backends = {'dummy': {'nq': 4, 'target': 'dummy'}}

    monkeypatch.setattr(
        _ionq_http_client.IonQ,
        'update_devices_list',
        _dummy_update.__get__(None, _ionq_http_client.IonQ),
    )

    mock_response = mock.MagicMock()
    mock_response.status_code = 400
    mock_response.json = mock.MagicMock(
        return_value={
            'error': 'Bad Request',
            'message': 'Invalid request body',
        }
    )
    auth_error = requests.exceptions.HTTPError(response=mock_response)
    mock_post = mock.MagicMock(side_effect=auth_error)
    monkeypatch.setattr('requests.sessions.Session.post', mock_post)

    def user_password_input(prompt):
        if prompt == 'IonQ apiKey > ':
            return 'NotNone'

    monkeypatch.setattr('getpass.getpass', user_password_input)

    # Code to test:
    info = {
        'nq': 1,
        'shots': 1,
        'meas_mapped': [],
        'meas_qubit_ids': [],
        'circuit': [],
    }
    with pytest.raises(JobSubmissionError) as excinfo:
        _ionq_http_client.send(info, device='dummy')
    mock_post.assert_called_once()
    assert str(excinfo.value) == "Bad Request: Invalid request body"


def test_send_auth_token_required(monkeypatch):
    # Patch the method to give back dummy devices
    def _dummy_update(_self):
        _self.backends = {'dummy': {'nq': 4, 'target': 'dummy'}}

    monkeypatch.setattr(
        _ionq_http_client.IonQ,
        'update_devices_list',
        _dummy_update.__get__(None, _ionq_http_client.IonQ),
    )

    mock_post = mock.MagicMock(side_effect=Exception())
    monkeypatch.setattr('requests.sessions.Session.post', mock_post)

    def user_password_input(prompt):
        if prompt == 'IonQ apiKey > ':
            return None

    monkeypatch.setattr('getpass.getpass', user_password_input)

    # Code to test:
    info = {
        'nq': 1,
        'shots': 1,
        'meas_mapped': [],
        'meas_qubit_ids': [],
        'circuit': [],
    }
    with pytest.raises(RuntimeError) as excinfo:
        _ionq_http_client.send(info, device='dummy')
    mock_post.assert_not_called()
    assert 'An authentication token is required!' == str(excinfo.value)


@pytest.mark.parametrize(
    "expected_err, err_data",
    [
        (
            "UnknownError: An unknown error occurred! (status=unknown)",
            {'status': 'unknown'},
        ),
        (
            'APIError: Something failed! (status=failed)',
            {
                'status': 'failed',
                'failure': {
                    'error': 'Something failed!',
                    'code': 'APIError',
                },
            },
        ),
    ],
)
def test_send_api_errors_are_raised(monkeypatch, expected_err, err_data):
    # Patch the method to give back dummy devices
    def _dummy_update(_self):
        _self.backends = {'dummy': {'nq': 4, 'target': 'dummy'}}

    monkeypatch.setattr(
        _ionq_http_client.IonQ,
        'update_devices_list',
        _dummy_update.__get__(None, _ionq_http_client.IonQ),
    )

    def mock_post(_self, path, **kwargs):
        assert _api_url[:-1] == path
        mock_response = mock.MagicMock()
        mock_response.json = mock.MagicMock(return_value=err_data)
        return mock_response

    monkeypatch.setattr('requests.sessions.Session.post', mock_post)

    def user_password_input(prompt):
        if prompt == 'IonQ apiKey > ':
            return 'NotNone'

    monkeypatch.setattr('getpass.getpass', user_password_input)

    # Code to test:
    info = {
        'nq': 1,
        'shots': 1,
        'meas_mapped': [],
        'meas_qubit_ids': [],
        'circuit': [],
    }
    with pytest.raises(JobSubmissionError) as excinfo:
        _ionq_http_client.send(info, device='dummy')

    assert expected_err == str(excinfo.value)


def test_timeout_exception(monkeypatch):
    # Patch the method to give back dummy devices
    def _dummy_update(_self):
        _self.backends = {'dummy': {'nq': 4, 'target': 'dummy'}}

    monkeypatch.setattr(
        _ionq_http_client.IonQ,
        'update_devices_list',
        _dummy_update.__get__(None, _ionq_http_client.IonQ),
    )

    def mock_post(_self, path, *args, **kwargs):
        assert path == _api_url[:-1]
        mock_response = mock.MagicMock()
        mock_response.json = mock.MagicMock(
            return_value={
                'id': 'new-job-id',
                'status': 'ready',
            }
        )
        return mock_response

    def mock_get(_self, path, *args, **kwargs):
        assert urljoin(_api_url, 'new-job-id') == path
        mock_response = mock.MagicMock()
        mock_response.json = mock.MagicMock(
            return_value={
                'id': 'new-job-id',
                'status': 'running',
            }
        )
        return mock_response

    monkeypatch.setattr('requests.sessions.Session.post', mock_post)
    monkeypatch.setattr('requests.sessions.Session.get', mock_get)

    def user_password_input(prompt):
        if prompt == 'IonQ apiKey > ':
            return 'NotNone'

    monkeypatch.setattr('getpass.getpass', user_password_input)

    # Called once per loop in _get_result while the job is not ready.
    mock_sleep = mock.MagicMock()
    monkeypatch.setattr(_ionq_http_client.time, 'sleep', mock_sleep)

    # RequestTimeoutErrors are not caught, and so will raise out.
    with pytest.raises(RequestTimeoutError) as excinfo:
        info = {
            'nq': 1,
            'shots': 1,
            'meas_mapped': [],
            'meas_qubit_ids': [],
            'circuit': [],
        }
        _ionq_http_client.send(info, device='dummy', num_retries=1)
    mock_sleep.assert_called_once()
    assert 'Timeout. The ID of your submitted job is new-job-id.' == str(excinfo.value)


@pytest.mark.parametrize('token', [None, 'NotNone'])
def test_retrieve(monkeypatch, token):
    def _dummy_update(_self):
        _self.backends = {'dummy': {'nq': 4, 'target': 'dummy'}}

    monkeypatch.setattr(
        _ionq_http_client.IonQ,
        'update_devices_list',
        _dummy_update.__get__(None, _ionq_http_client.IonQ),
    )
    request_num = [0]

    def mock_get(_self, path, *args, **kwargs):
        assert urljoin(_api_url, 'old-job-id') == path
        json_response = {
            'id': 'old-job-id',
            'status': 'running',
        }
        if request_num[0] > 1:
            json_response = {
                'id': 'old-job-id',
                'status': 'completed',
                'qubits': 4,
                'registers': {'meas_mapped': [2, 3]},
                'metadata': {'meas_qubit_ids': '[2, 3]'},
                'data': {
                    'registers': {'meas_mapped': {'2': 1}},
                },
            }
        mock_response = mock.MagicMock()
        mock_response.json = mock.MagicMock(return_value=json_response)
        request_num[0] += 1
        return mock_response

    monkeypatch.setattr('requests.sessions.Session.get', mock_get)

    def user_password_input(prompt):
        if prompt == 'IonQ apiKey > ':
            return 'NotNone'

    monkeypatch.setattr('getpass.getpass', user_password_input)

    expected = {
        'nq': 4,
        'output_probs': {'2': 1},
        'meas_qubit_ids': [2, 3],
        'meas_mapped': [2, 3],
    }

    # Code to test:
    # Called once per loop in _get_result while the job is not ready.
    mock_sleep = mock.MagicMock()
    monkeypatch.setattr(_ionq_http_client.time, 'sleep', mock_sleep)
    result = _ionq_http_client.retrieve('dummy', token, 'old-job-id')
    assert expected == result
    # We only sleep twice.
    assert 2 == mock_sleep.call_count


def test_retrieve_that_errors_are_caught(monkeypatch):
    def _dummy_update(_self):
        _self.backends = {'dummy': {'nq': 4, 'target': 'dummy'}}

    monkeypatch.setattr(
        _ionq_http_client.IonQ,
        'update_devices_list',
        _dummy_update.__get__(None, _ionq_http_client.IonQ),
    )
    request_num = [0]

    def mock_get(_self, path, *args, **kwargs):
        assert urljoin(_api_url, 'old-job-id') == path
        json_response = {
            'id': 'old-job-id',
            'status': 'running',
        }
        if request_num[0] > 0:
            json_response = {
                'id': 'old-job-id',
                'status': 'failed',
                'failure': {
                    'code': 'ErrorCode',
                    'error': 'A descriptive error message.',
                },
            }
        mock_response = mock.MagicMock()
        mock_response.json = mock.MagicMock(return_value=json_response)
        request_num[0] += 1
        return mock_response

    monkeypatch.setattr('requests.sessions.Session.get', mock_get)

    def user_password_input(prompt):
        if prompt == 'IonQ apiKey > ':
            return 'NotNone'

    monkeypatch.setattr('getpass.getpass', user_password_input)

    # Code to test:
    mock_sleep = mock.MagicMock()
    monkeypatch.setattr(_ionq_http_client.time, 'sleep', mock_sleep)
    with pytest.raises(Exception):
        _ionq_http_client.retrieve('dummy', 'NotNone', 'old-job-id')
    mock_sleep.assert_called_once()



# -*- coding: utf-8 -*-
#   Copyright 2020 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for projectq.libs.math._gates.py."""

from projectq.libs.math import (
    AddConstant,
    AddConstantModN,
    MultiplyByConstantModN,
    SubConstant,
    SubConstantModN,
    AddQuantum,
    SubtractQuantum,
    ComparatorQuantum,
    DivideQuantum,
    MultiplyQuantum,
)

from ._gates import (
    AddQuantumGate,
    SubtractQuantumGate,
    MultiplyQuantumGate,
    DivideQuantumGate,
    ComparatorQuantumGate,
)


def test_addconstant():
    assert AddConstant(3) == AddConstant(3)
    assert not AddConstant(3) == AddConstant(4)
    assert AddConstant(7) != AddConstant(3)
    assert str(AddConstant(3)) == "AddConstant(3)"


def test_addconstantmodn():
    assert AddConstantModN(3, 4) == AddConstantModN(3, 4)
    assert not AddConstantModN(3, 4) == AddConstantModN(4, 4)
    assert not AddConstantModN(3, 5) == AddConstantModN(3, 4)
    assert AddConstantModN(7, 4) != AddConstantModN(3, 4)
    assert AddConstantModN(3, 5) != AddConstantModN(3, 4)
    assert str(AddConstantModN(3, 4)) == "AddConstantModN(3, 4)"


def test_multiplybyconstmodn():
    assert MultiplyByConstantModN(3, 4) == MultiplyByConstantModN(3, 4)
    assert not MultiplyByConstantModN(3, 4) == MultiplyByConstantModN(4, 4)
    assert not MultiplyByConstantModN(3, 5) == MultiplyByConstantModN(3, 4)
    assert MultiplyByConstantModN(7, 4) != MultiplyByConstantModN(3, 4)
    assert MultiplyByConstantModN(3, 5) != MultiplyByConstantModN(3, 4)
    assert str(MultiplyByConstantModN(3, 4)) == "MultiplyByConstantModN(3, 4)"


def test_AddQuantum():
    assert AddQuantum == AddQuantumGate()
    assert AddQuantum != SubtractQuantum
    assert not AddQuantum == SubtractQuantum
    assert str(AddQuantum) == "AddQuantum"


def test_SubtractQuantum():
    assert SubtractQuantum == SubtractQuantumGate()
    assert SubtractQuantum != AddQuantum
    assert not SubtractQuantum == ComparatorQuantum
    assert str(SubtractQuantum) == "SubtractQuantum"


def test_Comparator():
    assert ComparatorQuantum == ComparatorQuantumGate()
    assert ComparatorQuantum != AddQuantum
    assert not ComparatorQuantum == AddQuantum
    assert str(ComparatorQuantum) == "Comparator"


def test_QuantumDivision():
    assert DivideQuantum == DivideQuantumGate()
    assert DivideQuantum != MultiplyQuantum
    assert not DivideQuantum == MultiplyQuantum
    assert str(DivideQuantum) == "DivideQuantum"


def test_QuantumMultiplication():
    assert MultiplyQuantum == MultiplyQuantumGate()
    assert MultiplyQuantum != DivideQuantum
    assert not MultiplyQuantum == DivideQuantum
    assert str(MultiplyQuantum) == "MultiplyQuantum"


def test_hash_function_implemented():
    assert hash(AddConstant(3)) == hash(str(AddConstant(3)))
    assert hash(SubConstant(-3)) == hash(str(AddConstant(3)))
    assert hash(AddConstantModN(7, 4)) == hash(str(AddConstantModN(7, 4)))
    assert hash(SubConstantModN(7, 4)) == hash(str(AddConstantModN(-3, 4)))
    assert hash(MultiplyByConstantModN(3, 5)) == hash(str(MultiplyByConstantModN(3, 5)))
    assert hash(AddQuantum) == hash(str(AddQuantum))
    assert hash(SubtractQuantum) == hash(str(SubtractQuantum))
    assert hash(ComparatorQuantum) == hash(str(ComparatorQuantum))
    assert hash(DivideQuantum) == hash(str(DivideQuantum))
    assert hash(MultiplyQuantum) == hash(str(MultiplyQuantum))



# -*- coding: utf-8 -*-
# pylint: skip-file

import matplotlib.pyplot as plt
import getpass

from projectq import MainEngine
from projectq.backends import IBMBackend
from projectq.libs.hist import histogram
from projectq.ops import Measure, Entangle, All
import projectq.setups.ibm


def run_entangle(eng, num_qubits=3):
    """
    Runs an entangling operation on the provided compiler engine.

    Args:
        eng (MainEngine): Main compiler engine to use.
        num_qubits (int): Number of qubits to entangle.

    Returns:
        measurement (list<int>): List of measurement outcomes.
    """
    # allocate the quantum register to entangle
    qureg = eng.allocate_qureg(num_qubits)

    # entangle the qureg
    Entangle | qureg

    # measure; should be all-0 or all-1
    All(Measure) | qureg

    # run the circuit
    eng.flush()

    # access the probabilities via the back-end:
    # results = eng.backend.get_probabilities(qureg)
    # for state in results:
    #     print("Measured {} with p = {}.".format(state, results[state]))
    # or plot them directly:
    histogram(eng.backend, qureg)
    plt.show()

    # return one (random) measurement outcome.
    return [int(q) for q in qureg]


if __name__ == "__main__":
    # devices commonly available :
    # ibmq_16_melbourne (15 qubit)
    # ibmq_essex (5 qubit)
    # ibmq_qasm_simulator (32 qubits)
    # and plenty of other 5 qubits devices!
    #
    # To get a token, create a profile at:
    # https://quantum-computing.ibm.com/
    #
    device = None  # replace by the IBM device name you want to use
    token = None  # replace by the token given by IBMQ
    if token is None:
        token = getpass.getpass(prompt='IBM Q token > ')
    if device is None:
        device = getpass.getpass(prompt='IBM device > ')
    # create main compiler engine for the IBM back-end
    eng = MainEngine(
        IBMBackend(use_hardware=True, token=token, num_runs=1024, verbose=False, device=device),
        engine_list=projectq.setups.ibm.get_engine_list(token=token, device=device),
    )
    # run the circuit and print the result
    print(run_entangle(eng))
